"""coBib's shell helpers.

This module provides a variety of shell helper utilities.
"""
# pylint: disable=unused-argument

import argparse
import contextlib
import inspect
import logging
from io import StringIO
from typing import List, Set

from .rel_path import RelPath


def list_commands(args: List[str]) -> List[str]:
    """Lists all available subcommands.

    Args:
        args: a sequence of additional arguments used for the execution. None are supported yet.

    Returns:
        The list of available commands.
    """
    # pylint: disable=import-outside-toplevel
    from cobib import commands

    return [cls.name for _, cls in inspect.getmembers(commands) if inspect.isclass(cls)]


def list_labels(args: List[str]) -> List[str]:
    """List all available labels in the database.

    Args:
        args: a sequence of additional arguments used for the execution. None are supported yet.

    Returns:
        The list of all labels.
    """
    # pylint: disable=import-outside-toplevel
    from cobib.database import Database

    labels = list(Database().keys())
    return labels


def list_filters(args: List[str]) -> Set[str]:
    """Lists all field names available for filtering.

    Args:
        args: a sequence of additional arguments used for the execution. None are supported yet.

    Returns:
        The list of all available filters.
    """
    # pylint: disable=import-outside-toplevel
    from cobib.database import Database

    filters: Set[str] = {"label"}
    for entry in Database().values():
        filters.update(entry.data.keys())
    return filters


def example_config(args: List[str]) -> List[str]:
    """Shows the (well-commented) example configuration.

    Args:
        args: a sequence of additional arguments used for the execution. None are supported yet.

    Returns:
        The lines of the example config file.
    """
    root = RelPath(__file__).parent.parent
    with open(root / "config/example.py", "r", encoding="utf-8") as file:
        return [line.strip() for line in file.readlines()]


class LintFormatter(logging.Formatter):
    """A custom logging.Formatter."""

    def __init__(self, *args, **kwargs) -> None:  # type: ignore
        # noqa: D107
        super().__init__(*args, **kwargs)

        self.dirty_entries: Set[str] = set()

        # pylint: disable=import-outside-toplevel
        from cobib.config import config

        self._database_path = RelPath(config.database.file)

        with open(self._database_path.path, "r", encoding="utf-8") as database:
            self._raw_database = database.readlines()

    def format(self, record: logging.LogRecord) -> str:
        """Format's the LogRecord.

        This custom Formatter uses the LogRecord's attributes to determine from which line of the
        raw database a formatting information was raised. The corresponding line number is used in
        conjunction with the actual message of the LogRecord for the formatting.

        Args:
            record: the LogRecord to be formatted.

        Returns:
            A string encoding the LogRecord's information.
        """
        try:
            entry = record.entry  # type: ignore[attr-defined]
            field = record.field  # type: ignore[attr-defined]
            self.dirty_entries.add(entry)
            raw_db = enumerate(self._raw_database)
            _, line = next(raw_db)
            while not line.startswith(entry):
                _, line = next(raw_db)
            while not line.strip().startswith(field):
                line_no, line = next(raw_db)
            return f"{self._database_path}:{line_no+1} {record.getMessage()}"
        except AttributeError:
            return ""


def lint_database(args: List[str]) -> List[str]:
    """Lints the users database.

    Args:
        args: a sequence of additional arguments used for the execution. Currently, only a single
            optional value is allowed:
                * `-f`, `--format`: if specified, the database will be formatted to automatically
                    resolve all lint messages.

    Returns:
        The list of INFO log messages raised upon Entry initialization.
    """
    parser = argparse.ArgumentParser(prog="lint_database", description="A database format linter.")
    parser.add_argument(
        "-f",
        "--format",
        action="store_true",
        help="Automatically format database to conform with linter.",
    )
    largs = parser.parse_args(args)

    # pylint: disable=import-outside-toplevel
    from cobib.database import Database

    output = StringIO()

    handler = logging.StreamHandler(output)
    handler.setLevel(logging.INFO)
    handler.addFilter(logging.Filter("cobib.database.entry"))

    formatter = LintFormatter()
    handler.setFormatter(formatter)

    root_logger = logging.getLogger()
    if root_logger.getEffectiveLevel() > logging.INFO:
        # overwriting all existing handlers with this local one
        root_logger.handlers = [handler]
        root_logger.setLevel(logging.INFO)
    else:
        # appending new handler
        root_logger.addHandler(handler)

    # trigger database reading to cause lint messages upon entry-construction
    Database.read()

    lint_messages = output.getvalue().split("\n")

    root_logger.removeHandler(handler)

    if all(not msg for msg in lint_messages):
        return ["Congratulations! Your database triggers no lint messages."]

    if largs.format:
        for label in formatter.dirty_entries:
            # we exploit the rename method to register all dirty entries for re-writing
            Database().rename(label, label)

        Database.save()

        # pylint: disable=import-outside-toplevel
        from cobib.commands.base_command import Command

        # generate automatic git commit
        class LintCommand(Command):
            """The linting command."""

            name = "lint"

            def execute(self, args, out=None):  # type: ignore
                pass

            @staticmethod
            def tui(tui):  # type: ignore
                pass

        LintCommand().git(args=vars(largs))

        return ["The following lint messages have successfully been resolved:"] + lint_messages

    return lint_messages


def unify_labels(args: List[str]) -> List[str]:
    """Unifies all labels of the database according to `config.database.format.label_default`.

    Without the `--apply` argument this will only print the modification which would be performed!

    Args:
        args: a sequence of additional arguments used for the execution. Currently, only a single
            optional value is allowed:
                * `-a`, `--apply`: if specified, the label unification will actually be applied. The
                    default is to run in "dry"-mode which only prints the modifications.

    Returns:
        The list of INFO log messages raised upon label unification.
    """
    parser = argparse.ArgumentParser(prog="unify_labels", description="Label unification")
    parser.add_argument(
        "-a",
        "--apply",
        action="store_true",
        help="Actually apply the modifications rather than run in 'dry'-mode",
    )
    largs = parser.parse_args(args)

    # pylint: disable=import-outside-toplevel
    from cobib.commands import ModifyCommand
    from cobib.config import config

    modify_args = [
        "--dry",
        f"label:{config.database.format.label_default}",
        "--",
        "--label",
        "some_non_existend_label",  # this ensures that the command gets run on the entire database
    ]
    if largs.apply:
        modify_args = modify_args[1:]

    with contextlib.redirect_stderr(StringIO()) as out:
        ModifyCommand().execute(modify_args)

    return out.getvalue().strip().split("\n")



"""Tests for coBib's TUI Frame."""

import re
from typing import Any, Generator, List, Optional, Set, Tuple, Union

import pytest

from cobib.config import config
from cobib.tui import TUI, Frame, Mode, State
from cobib.tui.buffer import TextBuffer
from cobib.tui.state import STATE

from .. import get_resource
from .mock_curses import MockCursesPad
from .mock_tui import MockTUI


class TestTextFrame:
    """Tests for coBib's Frame."""

    @pytest.fixture(autouse=True)
    def setup(self, monkeypatch: pytest.MonkeyPatch) -> Generator[Any, None, None]:
        """Setup.

        This fixture is automatically enabled for all tests in this class.

        Args:
            monkeypatch: the built-in pytest fixture.
        """
        # pylint: disable=attribute-defined-outside-init
        monkeypatch.setattr("curses.newpad", lambda *args: MockCursesPad())
        self.frame = Frame(MockTUI(), 10, 20)  # type: ignore
        yield
        STATE.reset()

    def test_clear(self) -> None:
        """Test `cobib.tui.frame.Frame.clear`."""
        self.frame.buffer.lines = ["test"]
        self.frame.clear()
        assert self.frame.buffer.lines == []
        assert self.frame.buffer.height == 0
        assert self.frame.buffer.width == 0
        assert self.frame.buffer.wrapped is False
        old_buffer, old_state = self.frame.history[0]
        assert old_buffer.lines == ["test"]
        assert isinstance(old_state, State)

    @pytest.mark.parametrize(
        ["lines", "selection", "mode"],
        [
            [["Label0  Title0 by Author0"], set(), Mode.LIST.value],
            [["\x1b[37;45mLabel0\x1b[0m  Title0 by Author0"], set(), Mode.LIST.value],
            [["\x1b[37;45mLabel0\x1b[0m  Title0 by Author0"], {"Label0"}, Mode.LIST.value],
            [
                ["\x1b[34;40mLabel0\x1b[0m - 1 match", "[1]    Some matched text."],
                set(),
                Mode.SEARCH.value,
            ],
            [
                [
                    "\x1b[37;45m\x1b[34;40mLabel0\x1b[0m\x1b[0m - 1 match",
                    "[1]    Some matched text.",
                ],
                set(),
                Mode.SEARCH.value,
            ],
            [
                [
                    "\x1b[37;45m\x1b[34;40mLabel0\x1b[0m\x1b[0m - 1 match",
                    "[1]    Some matched text.",
                ],
                {"Label0"},
                Mode.SEARCH.value,
            ],
            [["@misc{Label0,", "author = {Nobody.}}"], set(), Mode.SHOW.value],
            [["@misc{\x1b[37;45mLabel0\x1b[0m,", "author = {Nobody.}}"], set(), Mode.SHOW.value],
            [
                ["@misc{\x1b[37;45mLabel0\x1b[0m,", "author = {Nobody.}}"],
                {"Label0"},
                Mode.SHOW.value,
            ],
        ],
    )
    def test_revert(
        self, caplog: pytest.LogCaptureFixture, lines: List[str], selection: Set[str], mode: str
    ) -> None:
        """Test `cobib.tui.frame.Frame.revert`.

        Args:
            caplog: the built-in pytest fixture.
            lines: a list of strings to place into the frame's buffer.
            selection: the set of selected labels.
            mode: the `cobib.tui.state.Mode` value.
        """
        buffer = TextBuffer()
        buffer.lines = lines
        buffer.height = len(lines)
        state = State()
        state.mode = mode
        self.frame.tui.selection = selection
        self.frame.history = [(buffer, state)]
        self.frame.revert()
        # assert buffer contents
        text = "\n".join(self.frame.buffer.lines)
        for label in selection:
            assert f"\x1b[37;45m{label}\x1b[0m" in text
        highlighted = re.findall(re.escape("\x1b[37;45m") + r"(.+)" + re.escape("\x1b[0m"), text)
        assert set(re.sub(re.escape("\x1b[0m"), "", h) for h in highlighted) == selection
        # assert log contents
        expected_log = [
            ("MockCursesPad", 10, "erase"),
            ("MockCursesPad", 10, "refresh: 0 0 1 0 10 19"),
            ("MockCursesPad", 10, f"resize: {len(lines)+1} 20"),
            ("MockCursesPad", 10, "refresh: 0 0 1 0 10 19"),
        ]
        for idx, line in enumerate(lines):
            expected_log.insert(-1, ("MockCursesPad", 10, f"addstr: {idx} 0 {line}"))
        assert [
            record
            for record in caplog.record_tuples
            if record[0] in ("MockCursesPad", "cobib.tui.frame")
        ] == expected_log

    def test_revert_empty_history(self, caplog: pytest.LogCaptureFixture) -> None:
        """Test revert with empty history method.

        Args:
            caplog: the built-in pytest fixture.
        """
        self.frame.revert()
        expected_log = [("cobib.tui.frame", 10, "Empty frame history, nothing to revert")]
        assert [
            record for record in caplog.record_tuples if record[0] == "cobib.tui.frame"
        ] == expected_log

    def test_resize(self, caplog: pytest.LogCaptureFixture) -> None:
        """Test `cobib.tui.frame.Frame.resize`.

        Args:
            caplog: the built-in pytest fixture.
        """
        self.frame.resize(10, 10)
        expected_log = [
            ("MockCursesPad", 10, "refresh: 0 0 1 0 10 9"),
        ]
        assert [
            record for record in caplog.record_tuples if record[0] == "MockCursesPad"
        ] == expected_log

    def test_refresh(self, caplog: pytest.LogCaptureFixture) -> None:
        """Test `cobib.tui.frame.Frame.refresh`.

        Args:
            caplog: the built-in pytest fixture.
        """
        self.frame.height = 10
        self.frame.width = 10
        self.frame.refresh()
        expected_log = [
            ("MockCursesPad", 10, "refresh: 0 0 1 0 10 9"),
        ]
        assert [
            record for record in caplog.record_tuples if record[0] == "MockCursesPad"
        ] == expected_log

    def test_view(self, caplog: pytest.LogCaptureFixture) -> None:
        """Test `cobib.tui.frame.Frame.view`.

        Args:
            caplog: the built-in pytest fixture.
        """
        self.frame.buffer.lines = [
            "Label0  Title0 by Author0",
            "Label1  Title1 by Author1",
            "Label2  Title2 by Author2",
            "Label3  Title3 by Author3",
            "Label4  Title4 by Author4",
        ]
        self.frame.view()
        expected_log = [
            ("MockCursesPad", 10, "erase"),
            ("MockCursesPad", 10, "refresh: 0 0 1 0 10 19"),
            ("MockCursesPad", 10, "resize: 1 20"),
            ("MockCursesPad", 10, "addstr: 0 0 Label0  Title0 by Author0"),
            ("MockCursesPad", 10, "addstr: 1 0 Label1  Title1 by Author1"),
            ("MockCursesPad", 10, "addstr: 2 0 Label2  Title2 by Author2"),
            ("MockCursesPad", 10, "addstr: 3 0 Label3  Title3 by Author3"),
            ("MockCursesPad", 10, "addstr: 4 0 Label4  Title4 by Author4"),
            ("MockCursesPad", 10, "refresh: 0 0 1 0 10 19"),
        ]
        assert [
            record for record in caplog.record_tuples if record[0] == "MockCursesPad"
        ] == expected_log

    def test_view_with_ansi_map(
        self, caplog: pytest.LogCaptureFixture, monkeypatch: pytest.MonkeyPatch
    ) -> None:
        """Test view method with ANSI color map.

        Args:
            caplog: the built-in pytest fixture.
            monkeypatch: the built-in pytest fixture.
        """
        monkeypatch.setattr("curses.color_pair", lambda *args: args)
        # create ANSI color map for testing purposes
        config.defaults()
        ansi_map = {}
        for attr in TUI.COLOR_NAMES:
            ansi_map[config.get_ansi_color(attr)] = TUI.COLOR_NAMES.index(attr) + 1
        # populate buffer
        self.frame.buffer.lines = [
            "\x1b[31;40mLabel0\x1b[0m  Title0 by Author0",
            "\x1b[31;40mLabel1\x1b[0m  Title1 by Author1",
            "\x1b[31;40mLabel2\x1b[0m  Title2 by Author2",
            "\x1b[31;40mLabel3\x1b[0m  Title3 by Author3",
            "\x1b[31;40mLabel4\x1b[0m  Title4 by Author4",
        ]
        self.frame.view(ansi_map=ansi_map)
        expected_log = [
            ("MockCursesPad", 10, "erase"),
            ("MockCursesPad", 10, "refresh: 0 0 1 0 10 19"),
            ("MockCursesPad", 10, "resize: 1 20"),
            ("MockCursesPad", 10, "addstr: 0 0 Label0  Title0 by Author0"),
            ("MockCursesPad", 10, "chgat: 0 0 6 (4,)"),
            ("MockCursesPad", 10, "addstr: 1 0 Label1  Title1 by Author1"),
            ("MockCursesPad", 10, "chgat: 1 0 6 (4,)"),
            ("MockCursesPad", 10, "addstr: 2 0 Label2  Title2 by Author2"),
            ("MockCursesPad", 10, "chgat: 2 0 6 (4,)"),
            ("MockCursesPad", 10, "addstr: 3 0 Label3  Title3 by Author3"),
            ("MockCursesPad", 10, "chgat: 3 0 6 (4,)"),
            ("MockCursesPad", 10, "addstr: 4 0 Label4  Title4 by Author4"),
            ("MockCursesPad", 10, "chgat: 4 0 6 (4,)"),
            ("MockCursesPad", 10, "refresh: 0 0 1 0 10 19"),
        ]
        assert [
            record for record in caplog.record_tuples if record[0] == "MockCursesPad"
        ] == expected_log
        for message in ["Interpreting ANSI color codes on the fly.", "Applying ANSI color map."]:
            assert ("cobib.tui.buffer", 10, message) in caplog.record_tuples

    @pytest.mark.parametrize(
        ["initialize", "update", "repeat", "top", "cur", "msg", "scrolloff"],
        [
            [(0, 0), "g", 1, 0, 0, "Jump to top of viewport.", None],
            [(0, 0), "G", 1, 10, 19, "Jump to bottom of viewport.", None],
            [(0, 0), 1, 1, 0, 1, "Scroll viewport down by 1 lines.", None],
            [(0, 0), 1, 10, 4, 10, "Scroll viewport down by 1 lines.", None],
            [(0, 0), 10, 1, 10, 10, "Scroll viewport down by 10 lines.", None],
            [(0, 0), 20, 1, 10, 19, "Scroll viewport down by 20 lines.", None],
            [(0, 0), 1, 10, 5, 10, "Scroll viewport down by 1 lines.", 10],
            [(0, 0), 10, 1, 5, 10, "Scroll viewport down by 10 lines.", 10],
            [(0, 5), -1, 1, 0, 4, "Scroll viewport up by -1 lines.", None],
            [(0, 5), -1, 10, 0, 0, "Scroll viewport up by -1 lines.", None],
            [(0, 5), -10, 1, 0, 0, "Scroll viewport up by -10 lines.", None],
            [(10, 19), -1, 10, 6, 9, "Scroll viewport up by -1 lines.", None],
            [(10, 19), -10, 1, 0, 9, "Scroll viewport up by -10 lines.", None],
            [(10, 19), -20, 1, 0, 0, "Scroll viewport up by -20 lines.", None],
            [(10, 19), -1, 10, 4, 9, "Scroll viewport up by -1 lines.", 10],
            [(10, 19), -10, 1, 4, 9, "Scroll viewport up by -10 lines.", 10],
            [(10, 19), -20, 1, 0, 0, "Scroll viewport up by -20 lines.", 10],
        ],
    )
    # when `scrolloff` is None, the default will be used
    def test_scroll_y(
        self,
        caplog: pytest.LogCaptureFixture,
        initialize: Tuple[int, int],
        update: Union[int, str],
        repeat: int,
        top: int,
        cur: int,
        msg: str,
        scrolloff: Optional[int],
    ) -> None:
        """Test `cobib.tui.frame.Frame.scroll_y`.

        Args:
            caplog: the built-in pytest fixture.
            initialize: the initial values for `cobib.tui.state.State.top_line` and
                `cobib.tui.state.State.current_line`.
            update: the argument to send to `cobib.tui.frame.Frame.scroll_y`.
            repeat: the number of times to repeat the update.
            top: the expected value for `cobib.tui.state.State.top_line`.
            cur: the expected value for `cobib.tui.state.State.current_line`.
            msg: the expected message to be logged.
            scrolloff: the configuration value for `config.tui.scroll_offset`.
        """
        config.defaults()
        if scrolloff:
            config.tui.scroll_offset = scrolloff

        self.frame.buffer.height = 20
        self.frame.buffer.lines = ["Many lines" for _ in range(20)]

        STATE.reset()
        STATE.top_line = initialize[0]
        STATE.current_line = initialize[1]

        for _ in range(repeat):
            self.frame.scroll_y(update)
            assert ("cobib.tui.frame", 10, msg) in caplog.record_tuples
        assert STATE.top_line == top
        assert STATE.current_line == cur

    @pytest.mark.parametrize(
        ["initialize", "update", "repeat", "left", "msg"],
        [
            [0, 0, 1, 0, "Jump to first column of viewport."],
            [0, "$", 1, 70, "Jump to end of viewport."],
            [0, 1, 1, 1, "Scroll viewport horizontally by 1 columns."],
            [0, 1, 69, 69, "Scroll viewport horizontally by 1 columns."],
            [0, 1, 70, 70, "Scroll viewport horizontally by 1 columns."],
            [0, 1, 71, 70, "Scroll viewport horizontally by 1 columns."],
            [0, 10, 1, 10, "Scroll viewport horizontally by 10 columns."],
            [0, 70, 1, 70, "Scroll viewport horizontally by 70 columns."],
            [10, -1, 1, 9, "Scroll viewport horizontally by -1 columns."],
            [10, -1, 10, 0, "Scroll viewport horizontally by -1 columns."],
            [70, -1, 69, 1, "Scroll viewport horizontally by -1 columns."],
            [70, -1, 70, 0, "Scroll viewport horizontally by -1 columns."],
            [70, -1, 71, 0, "Scroll viewport horizontally by -1 columns."],
            [20, -10, 1, 10, "Scroll viewport horizontally by -10 columns."],
            [70, -70, 1, 0, "Scroll viewport horizontally by -70 columns."],
        ],
    )
    def test_scroll_x(
        self,
        caplog: pytest.LogCaptureFixture,
        initialize: int,
        update: Union[int, str],
        repeat: int,
        left: int,
        msg: str,
    ) -> None:
        """Test `cobib.tui.frame.Frame.scroll_x`.

        Args:
            caplog: the built-in pytest fixture.
            initialize: the initial value for `cobib.tui.state.State.left_edge`.
            update: the argument to send to `cobib.tui.frame.Frame.scroll_x`.
            repeat: the number of times to repeat the update.
            left: the expected value for `cobib.tui.state.State.left_edge`.
            msg: the expected message to be logged.
        """
        text = "A very long line. " * 5
        self.frame.buffer.lines = [text]
        self.frame.buffer.width = len(text)

        STATE.reset()
        STATE.left_edge = initialize

        for _ in range(repeat):
            self.frame.scroll_x(update)
            assert ("cobib.tui.frame", 10, msg) in caplog.record_tuples
        assert STATE.left_edge == left

    def test_wrap(self) -> None:
        """Test `cobib.tui.frame.Frame.wrap`."""
        self.frame.buffer.lines = ["Long line with some text" for _ in range(10)]
        self.frame.wrap()
        assert STATE.left_edge == 0
        STATE.current_line = 20
        self.frame.wrap()
        assert STATE.current_line == 9

    @pytest.mark.parametrize(
        ["lines", "cur_y", "mode"],
        [
            [["Label0  Title0 by Author0", "Label1  Title1 by Author1"], 0, Mode.LIST.value],
            [["Label0  Title0 by", "↪ Author0", "Label1  Title1 by Author1"], 1, Mode.LIST.value],
            [
                ["\x1b[34;40mLabel0\x1b[0m - 1 match", "[1]    Some matched text."],
                0,
                Mode.SEARCH.value,
            ],
            [
                ["\x1b[34;40mLabel0\x1b[0m - 1 match", "[1]    Some matched text."],
                1,
                Mode.SEARCH.value,
            ],
            [
                ["\x1b[34;40mLabel0\x1b[0m - 1 match", "[1]    Some matched", "↪ text."],
                2,
                Mode.SEARCH.value,
            ],
        ],
    )
    def test_get_current_label(
        self,
        caplog: pytest.LogCaptureFixture,
        lines: List[str],
        cur_y: int,
        mode: str,
    ) -> None:
        """Test `cobib.tui.frame.Frame.get_current_label`.

        Args:
            caplog: the built-in pytest fixture.
            lines: a list of strings to place into the frame's buffer.
            cur_y: the value for `cobib.tui.state.State.current_line`.
            mode: the `cobib.tui.state.Mode` value.
        """
        self.frame.pad.lines = lines  # type: ignore
        self.frame.pad.current_pos[0] = cur_y  # type: ignore
        STATE.mode = mode
        cur_label, label_y = self.frame.get_current_label()
        assert cur_label == "Label0"
        assert label_y == 0
        expected_log = [
            ("cobib.tui.frame", 10, 'Obtaining current label "under" cursor.'),
            ("MockCursesPad", 10, "getyx"),
            ("MockCursesPad", 10, "instr: 0, 0"),
            ("cobib.tui.frame", 10, 'Current label at "0" is "Label0".'),
        ]
        for i in range(cur_y + 1):
            expected_log.insert(2, ("MockCursesPad", 10, f"inch: {i}, 0"))
        assert [
            record
            for record in caplog.record_tuples
            if record[0] in ("MockCursesPad", "cobib.tui.frame")
        ] == expected_log

    @pytest.mark.parametrize(
        ["lines", "cur_y", "topstatus"],
        [
            [["@misc{Label0,", "author = {Nobody.}}"], 0, "coBib version - Label0"],
            [["@misc{Label0,", "author = {Nobody.}}"], 1, "coBib version - Label0"],
        ],
    )
    def test_get_current_label_show(
        self,
        caplog: pytest.LogCaptureFixture,
        lines: List[str],
        cur_y: int,
        topstatus: str,
    ) -> None:
        """Test `cobib.tui.frame.Frame.get_current_label` for `cobib.tui.state.Mode.SHOW`.

        This is separate from `test_get_current_label` because the assertions are fundamentally
        different.

        Args:
            caplog: the built-in pytest fixture.
            lines: a list of strings to place into the frame's buffer.
            cur_y: the value for `cobib.tui.state.State.current_line`.
            topstatus: the value for `cobib.tui.state.State.topstatus`.
        """
        self.frame.pad.lines = lines  # type: ignore
        self.frame.pad.current_pos[0] = cur_y  # type: ignore
        STATE.topstatus = topstatus
        STATE.mode = Mode.SHOW.value
        cur_label, label_y = self.frame.get_current_label()
        assert cur_label == "Label0"
        assert label_y == 0
        expected_log = [
            ("cobib.tui.frame", 10, 'Obtaining current label "under" cursor.'),
            ("MockCursesPad", 10, "getyx"),
            ("cobib.tui.frame", 10, 'Current label at "0" is "Label0".'),
        ]
        assert [
            record
            for record in caplog.record_tuples
            if record[0] in ("MockCursesPad", "cobib.tui.frame")
        ] == expected_log

    @pytest.mark.parametrize(
        ["selection", "mode"],
        [
            [set(), Mode.LIST.value],
            [{"knuthwebsite"}, Mode.LIST.value],
            [{"knuthwebsite", "latexcompanion"}, Mode.LIST.value],
            [set(), Mode.SEARCH.value],
            [{"knuthwebsite"}, Mode.SEARCH.value],
            [{"knuthwebsite", "latexcompanion"}, Mode.SEARCH.value],
            [set(), Mode.SHOW.value],
            [{"knuthwebsite"}, Mode.SHOW.value],
            [{"knuthwebsite", "latexcompanion"}, Mode.SHOW.value],
        ],
    )
    def test_update_list(
        self, caplog: pytest.LogCaptureFixture, selection: Set[str], mode: str
    ) -> None:
        """Test `cobib.tui.frame.Frame.update_list`.

        Args:
            caplog: the built-in pytest fixture.
            selection: the set of selected labels.
            mode: the `cobib.tui.state.Mode` value.
        """
        # load testing config
        config.load(get_resource("debug.py"))
        # trigger list update
        STATE.mode = mode
        self.frame.tui.selection = selection
        self.frame.update_list()
        # assert state
        assert STATE.mode == Mode.LIST.value
        assert STATE.current_line == 0
        assert STATE.top_line == 0
        assert STATE.left_edge == 0
        assert STATE.inactive_commands == []
        # assert buffer contents
        text = "\n".join(self.frame.buffer.lines)
        for label in selection:
            assert f"\x1b[37;45m{label}\x1b[0m" in text
        highlighted = re.findall(re.escape("\x1b[37;45m") + r"(.+)" + re.escape("\x1b[0m"), text)
        assert set(re.sub(re.escape("\x1b[0m"), "", h) for h in highlighted) == selection
        # assert log contents
        expected_log = [
            ("cobib.tui.frame", 10, "Re-populating the viewport with the list command."),
            ("MockCursesPad", 10, "erase"),
            ("MockCursesPad", 10, "refresh: 0 0 1 0 10 19"),
            ("MockCursesPad", 10, "resize: 4 54"),
            (
                "MockCursesPad",
                10,
                "addstr: 0 0 "
                + (
                    "\x1b[37;45mknuthwebsite\x1b[0m"
                    if "knuthwebsite" in selection
                    else "knuthwebsite"
                )
                + "    Knuth: Computers and Typesetting      ",
            ),
            (
                "MockCursesPad",
                10,
                "addstr: 1 0 "
                + (
                    "\x1b[37;45mlatexcompanion\x1b[0m"
                    if "latexcompanion" in selection
                    else "latexcompanion"
                )
                + r"  The \LaTeX\ Companion                 ",
            ),
            (
                "MockCursesPad",
                10,
                r"addstr: 2 0 "
                + ("\x1b[37;45meinstein\x1b[0m" if "einstein" in selection else "einstein")
                + r"        Zur Elektrodynamik bewegter K{\"o}rper",
            ),
            ("MockCursesPad", 10, "refresh: 0 0 1 0 10 19"),
        ]
        assert [
            record
            for record in caplog.record_tuples
            if record[0] in ("MockCursesPad", "cobib.tui.frame")
        ] == expected_log

    def test_update_list_out_of_view(self) -> None:
        """Test `cobib.tui.frame.Frame.update_list` when the current line is out of view."""
        # load testing config
        config.load(get_resource("debug.py"))
        # make current line be out-of-view
        self.frame.height = 2
        STATE.current_line = 3
        # trigger list update
        self.frame.update_list()
        # assert state
        assert STATE.mode == Mode.LIST.value
        assert STATE.current_line == 3
        assert STATE.top_line == 1
        assert STATE.left_edge == 0
        assert STATE.inactive_commands == []



"""Tests for coBib's AddCommand."""
# pylint: disable=no-self-use,unused-argument

from __future__ import annotations

import os
import tempfile
from argparse import Namespace
from itertools import zip_longest
from typing import TYPE_CHECKING, Any, Dict, List, Optional, Type

import pytest

from cobib.commands import AddCommand
from cobib.config import Event, config
from cobib.database import Database, Entry
from cobib.utils.rel_path import RelPath

from .. import get_resource
from ..tui.tui_test import TUITest
from .command_test import CommandTest

EXAMPLE_LITERATURE = get_resource("example_literature.yaml")
EXAMPLE_DUPLICATE_ENTRY_BIB = get_resource("example_duplicate_entry.bib", "commands")
EXAMPLE_DUPLICATE_ENTRY_YAML = get_resource("example_duplicate_entry.yaml", "commands")
EXAMPLE_MULTI_FILE_ENTRY_BIB = get_resource("example_multi_file_entry.bib", "commands")
EXAMPLE_MULTI_FILE_ENTRY_YAML = get_resource("example_multi_file_entry.yaml", "commands")

if TYPE_CHECKING:
    import cobib.commands


class TestAddCommand(CommandTest, TUITest):
    """Tests for coBib's AddCommand."""

    def get_command(self) -> Type[cobib.commands.base_command.Command]:
        # noqa: D102
        return AddCommand

    def _assert(self, extra_filename: str) -> None:
        """Common assertion utility method.

        Args:
            extra_filename: path to an additional filename whose contents are to be added to the
                expected lines.
        """
        # compare with reference file
        with open(EXAMPLE_LITERATURE, "r", encoding="utf-8") as expected:
            true_lines = expected.readlines()
        with open(extra_filename, "r", encoding="utf-8") as extra:
            true_lines += extra.readlines()
        with open(config.database.file, "r", encoding="utf-8") as file:
            # we use zip_longest to ensure that we don't have more than we expect
            for line, truth in zip_longest(file, true_lines):
                assert line == truth

    def _assert_entry(self, label: str, **kwargs) -> None:  # type: ignore
        """An additional assertion utility to check specific entry fields.

        Args:
            label: the label of the entry.
            kwargs: additional keyword arguments whose contents are checked against the Entry's
                `data contents.
        """
        entry = Database()[label]
        for key, value in kwargs.items():
            assert entry.data.get(key, None) == value

    @pytest.mark.parametrize(
        ["setup"],
        [
            [{"git": False}],
            [{"git": True}],
        ],
        indirect=["setup"],
    )
    @pytest.mark.parametrize(
        ["more_args", "entry_kwargs"],
        [
            [[], {}],
            [
                ["-f", "test/debug.py"],
                {"file": [str(RelPath("test/debug.py"))]},
            ],
            [["-l", "dummy_label"], {}],
            [["tag"], {"tags": ["tag"]}],
            [["tag", "tag2"], {"tags": ["tag", "tag2"]}],
        ],
    )
    def test_command(self, setup: Any, more_args: List[str], entry_kwargs: Dict[str, Any]) -> None:
        """Test the command itself.

        Args:
            setup: the `tests.commands.command_test.CommandTest.setup` fixture.
            more_args: additional arguments to be passed to the command.
            entry_kwargs: the expected contents of the resulting `Entry`.
        """
        git = setup.get("git", False)

        try:
            label = more_args[more_args.index("-l") + 1]
        except ValueError:
            label = "example_multi_file_entry"
        args = ["-b", EXAMPLE_MULTI_FILE_ENTRY_BIB] + more_args

        AddCommand().execute(args)

        assert Database()[label]

        if entry_kwargs or label != "example_multi_file_entry":
            self._assert_entry(label, **entry_kwargs)
        else:
            # only when we don't use extra arguments the files will match
            self._assert(EXAMPLE_MULTI_FILE_ENTRY_YAML)

        if git:
            # assert the git commit message
            # Note: we do not assert the arguments, because they depend on the available parsers
            self.assert_git_commit_message("add", None)

    def test_add_new_entry(self, setup: Any, caplog: pytest.LogCaptureFixture) -> None:
        """Test adding a new plain entry.

        Args:
            setup: the `tests.commands.command_test.CommandTest.setup` fixture.
            caplog: the built-in pytest fixture.
        """
        AddCommand().execute(["-l", "dummy"])
        assert (
            "cobib.commands.add",
            30,
            "No input to parse. Creating new entry 'dummy' manually.",
        ) in caplog.record_tuples

        with open(config.database.file, "r", encoding="utf-8") as file:
            lines = file.readlines()
            dummy_start = lines.index("dummy:\n")
            assert dummy_start > 0
            assert lines[dummy_start - 1] == "---\n"
            assert lines[dummy_start + 1] == "  ENTRYTYPE: article\n"
            assert lines[dummy_start + 2] == "...\n"

    @pytest.mark.parametrize("folder", [None, "."])
    def test_add_with_download(
        self, folder: Optional[str], setup: Any, capsys: pytest.CaptureFixture[str]
    ) -> None:
        """Test adding a new entry with an associated file automatically downloaded.

        Args:
            folder: the folder for the downloaded file.
            setup: the `tests.commands.command_test.CommandTest.setup` fixture.
            capsys: the built-in pytest fixture.
        """
        path = RelPath(f"{'/tmp' if folder is None else folder}/Cao2018.pdf")
        try:
            # ensure file does not exist yet
            os.remove(path.path)
        except FileNotFoundError:
            pass
        try:
            args = ["-a", "1812.09976"]
            if folder:
                args += ["-p", folder]
            AddCommand().execute(args)
            entry = Database()["Cao2018"]
            assert entry.label == "Cao2018"
            assert entry.data["archivePrefix"] == "arXiv"
            assert entry.data["arxivid"].startswith("1812.09976")
            assert "_download" not in entry.data.keys()
            assert f"Successfully downloaded {path}" in capsys.readouterr().out
            assert os.path.exists(path.path)
        finally:
            try:
                os.remove(path.path)
            except FileNotFoundError:
                pass

    def test_add_skip_download(self, setup: Any) -> None:
        """Test adding a new entry and skipping the automatic download.

        Args:
            setup: the `tests.commands.command_test.CommandTest.setup` fixture.
        """
        path = RelPath("/tmp/Cao2018.pdf")
        try:
            args = ["-a", "1812.09976", "--skip-download"]
            AddCommand().execute(args)
            entry = Database()["Cao2018"]
            assert entry.label == "Cao2018"
            assert entry.data["archivePrefix"] == "arXiv"
            assert entry.data["arxivid"].startswith("1812.09976")
            assert "_download" not in entry.data.keys()
            assert not os.path.exists(path.path)
        finally:
            try:
                os.remove(path.path)
            except FileNotFoundError:
                pass

    @pytest.mark.parametrize(
        ["setup"],
        [
            [{"git": False}],
            [{"git": True}],
        ],
        indirect=["setup"],
    )
    def test_add_with_update(self, setup: Any) -> None:
        """Test update option of AddCommand.

        Args:
            setup: the `tests.commands.command_test.CommandTest.setup` fixture.
        """
        git = setup.get("git", False)
        AddCommand().execute(["-a", "1812.09976", "--skip-download"])

        # assert initial state
        entry = Database()["Cao2018"]

        assert entry.data["author"].startswith("Yudong Cao")
        assert entry.data["title"].startswith("Quantum Chemistry in the Age of Quantum Computing")
        assert entry.data["arxivid"].startswith("1812.09976")
        assert entry.data["doi"] == "10.1021/acs.chemrev.8b00803"
        assert entry.data["primaryClass"] == "quant-ph"
        assert entry.data["archivePrefix"] == "arXiv"
        assert entry.data["abstract"] != ""
        assert entry.data["year"] == 2018

        assert "journal" not in entry.data.keys()
        assert "month" not in entry.data.keys()
        assert "number" not in entry.data.keys()
        assert "pages" not in entry.data.keys()
        assert "volume" not in entry.data.keys()

        args = ["-d", "10.1021/acs.chemrev.8b00803", "-l", "Cao2018", "--skip-download", "--update"]
        AddCommand().execute(args)

        # assert final state
        entry = Database()["Cao2018"]

        assert entry.data["author"].startswith("Yudong Cao")
        assert entry.data["title"].startswith("Quantum Chemistry in the Age of Quantum Computing")
        assert entry.data["arxivid"].startswith("1812.09976")
        assert entry.data["primaryClass"] == "quant-ph"
        assert entry.data["archivePrefix"] == "arXiv"
        assert entry.data["abstract"] != ""

        assert entry.data["journal"] == "Chemical Reviews"
        assert entry.data["doi"] == "10.1021/acs.chemrev.8b00803"
        assert entry.data["month"] == "aug"
        assert entry.data["number"] == 19
        assert entry.data["pages"] == "10856--10915"
        assert entry.data["volume"] == 119
        assert entry.data["year"] == 2019

        if git:
            # assert the git commit message
            # Note: we do not assert the arguments, because they depend on the available parsers
            self.assert_git_commit_message("add", None)

    def test_skip_manual_add_if_exists(self, setup: Any, caplog: pytest.LogCaptureFixture) -> None:
        """Test manual addition is skipped if the label exists already.

        Args:
            setup: the `tests.commands.command_test.CommandTest.setup` fixture.
            caplog: the built-in pytest fixture.
        """
        AddCommand().execute(["-l", "einstein"])
        assert (
            "cobib.commands.add",
            30,
            "You tried to add a new entry 'einstein' which already exists!",
        ) in caplog.record_tuples
        assert (
            "cobib.commands.add",
            30,
            "Please use `cobib edit einstein` instead!",
        ) in caplog.record_tuples

    def test_continue_after_skip_exists(self, setup: Any, caplog: pytest.LogCaptureFixture) -> None:
        """Test entry addition continues after skipping over existing entry.

        Regression test against #83

        Args:
            setup: the `tests.commands.command_test.CommandTest.setup` fixture.
            caplog: the built-in pytest fixture.
        """
        with tempfile.NamedTemporaryFile("w") as file:
            with open(EXAMPLE_DUPLICATE_ENTRY_BIB, "r", encoding="utf-8") as existing:
                file.writelines(existing.readlines())
            file.writelines(["@article{dummy,\nauthor = {Dummy},\n}"])
            file.flush()
            AddCommand().execute(["--skip-existing", "-b", file.name])
        assert (
            "cobib.commands.add",
            30,
            "You tried to add a new entry 'einstein' which already exists!",
        ) in caplog.record_tuples
        assert (
            "cobib.commands.add",
            30,
            "Please use `cobib edit einstein` instead!",
        ) in caplog.record_tuples
        assert (
            "cobib.database.database",
            10,
            "Updating entry dummy",
        ) in caplog.record_tuples

    def test_warning_missing_label(self, setup: Any, caplog: pytest.LogCaptureFixture) -> None:
        """Test warning for missing label and any other input.

        Args:
            setup: the `tests.commands.command_test.CommandTest.setup` fixture.
            caplog: the built-in pytest fixture.
        """
        AddCommand().execute([""])
        assert (
            "cobib.commands.add",
            40,
            "Neither an input to parse nor a label for manual creation specified!",
        ) in caplog.record_tuples

    @pytest.mark.parametrize(
        ["setup"],
        [
            [{"git": False}],
            [{"git": True}],
        ],
        indirect=["setup"],
    )
    def test_overwrite_label(self, setup: Any) -> None:
        """Test add command while specifying a label manually.

        Regression test against #4.

        The duplicate entry has been adapted to also assert the elongation of Journal names.

        Args:
            setup: the `tests.commands.command_test.CommandTest.setup` fixture.
        """
        config.utils.journal_abbreviations = [("Annalen der Physik", "Ann. Phys.")]
        git = setup.get("git", False)
        # add potentially duplicate entry
        AddCommand().execute(["-b", EXAMPLE_DUPLICATE_ENTRY_BIB, "--label", "duplicate_resolver"])

        assert Database()["duplicate_resolver"]

        self._assert(EXAMPLE_DUPLICATE_ENTRY_YAML)

        if git:
            # assert the git commit message
            self.assert_git_commit_message("add", None)

    @pytest.mark.parametrize(
        ["setup"],
        [
            [{"git": False}],
            [{"git": True}],
        ],
        indirect=["setup"],
    )
    def test_configured_label_default(self, setup: Any) -> None:
        """Test add command when a `label_default` is pre-configured.

        Args:
            setup: the `tests.commands.command_test.CommandTest.setup` fixture.
        """
        config.database.format.label_default = "{author.split()[1]}{year}"
        git = setup.get("git", False)

        AddCommand().execute(["-b", EXAMPLE_DUPLICATE_ENTRY_BIB])

        assert Database()["Einstein1905"]

        if git:
            # assert the git commit message
            self.assert_git_commit_message("add", None)

    @pytest.mark.parametrize(
        ["setup"],
        [
            [{"git": False}],
            [{"git": True}],
        ],
        indirect=["setup"],
    )
    def test_disambiguate_label(self, setup: Any, caplog: pytest.LogCaptureFixture) -> None:
        """Test label disambiguation if the provided one already exists.

        Args:
            setup: the `tests.commands.command_test.CommandTest.setup` fixture.
            caplog: the built-in pytest fixture.
        """
        git = setup.get("git", False)

        AddCommand().execute(["-b", EXAMPLE_DUPLICATE_ENTRY_BIB])

        assert (
            "cobib.commands.add",
            30,
            "You tried to add a new entry 'einstein' which already exists!",
        ) in caplog.record_tuples
        assert (
            "cobib.commands.add",
            30,
            "The label will be disambiguated based on the configuration option: "
            "config.database.format.label_suffix",
        ) in caplog.record_tuples

        assert Database()["einstein_a"]

        if git:
            # assert the git commit message
            self.assert_git_commit_message("add", None)

    @pytest.mark.parametrize(
        ["setup"],
        [
            [{"git": False}],
        ],
        indirect=["setup"],
    )
    # other variants are already covered by test_command
    def test_cmdline(self, setup: Any, monkeypatch: pytest.MonkeyPatch) -> None:
        """Test the command-line access of the command.

        Args:
            setup: the `tests.commands.command_test.CommandTest.setup` fixture.
            monkeypatch: the built-in pytest fixture.
        """
        self.run_module(monkeypatch, "main", ["cobib", "add", "-b", EXAMPLE_MULTI_FILE_ENTRY_BIB])
        self._assert(EXAMPLE_MULTI_FILE_ENTRY_YAML)

    def test_tui(self, setup: Any) -> None:
        """Test the TUI access of the command.

        Args:
            setup: the `tests.commands.command_test.CommandTest.setup` fixture.
        """

        def assertion(screen, logs, **kwargs):  # type: ignore
            self._assert(EXAMPLE_MULTI_FILE_ENTRY_YAML)

            assert "example_multi_file_entry" in screen.display[1]

            expected_log = [
                ("cobib.commands.add", 10, "Add command triggered from TUI."),
                ("cobib.commands.add", 10, "Starting Add command."),
                (
                    "cobib.commands.add",
                    10,
                    "Adding entries from bibtex: '" + EXAMPLE_MULTI_FILE_ENTRY_BIB + "'.",
                ),
                ("cobib.commands.add", 20, "'example_multi_file_entry' was added to the database."),
                ("cobib.commands.add", 10, "Updating list after Add command."),
            ]
            assert [log for log in logs if log[0] == "cobib.commands.add"] == expected_log

        keys = "a-b " + EXAMPLE_MULTI_FILE_ENTRY_BIB + "\n"
        self.run_tui(keys, assertion, {})

    def test_event_pre_add_command(self, setup: Any) -> None:
        """Tests the PreAddCommand event."""

        @Event.PreAddCommand.subscribe
        def hook(largs: Namespace) -> None:
            largs.label = "dummy"

        assert Event.PreAddCommand.validate()

        AddCommand().execute(["-b", EXAMPLE_DUPLICATE_ENTRY_BIB])

        assert "dummy" in Database().keys()

    def test_event_post_add_command(self, setup: Any) -> None:
        """Tests the PostAddCommand event."""

        @Event.PostAddCommand.subscribe
        def hook(new_entries: Dict[str, Entry]) -> None:
            new_entries["dummy"] = new_entries.pop("einstein_a")

        assert Event.PostAddCommand.validate()

        AddCommand().execute(["-b", EXAMPLE_DUPLICATE_ENTRY_BIB])

        assert "dummy" in Database().keys()



"""coBib's curses-based TUI.

This class implements a curses-based TUI.
As such it provides an interactive front-end to coBib which allows easy navigation and manipulation
of the database.
It also benefits from some performance advantages over the command-line interface when doing many
operations because the parsed database entries remain in memory all the time and do not need to be
re-parsed upon every command invocation.

### Usage

The TUI exposes all commands through single-character key codes as well as some additional
TUI-specific commands:
* `Help` (defaults to the `?` key): opens a help popup.
* `Quit` (defaults to the `q` key): quits the TUI (or one level such as a popup).
* `Prompt` (defaults to the `:` key): starts the command prompt where a used can execute *any* coBib
  command-line command.
* `Select` (defaults to the `v` key): visually selects an entry.
* `Wrap` (defaults to the `w` key): wraps the buffer contents to the terminal width.

Additionally, the following navigation keys are available (these will be familiar to Vim-users):
* *arrow down*, `j`: move one line down
* *arrow up*, `k`: move one line up
* *arrow right*, `l`: move visible area one column to the right (has no effect when wrapped)
* *arrow left*, `h`: move visible area one column to the left (has no effect when wrapped)
* *page down*, `C-F`: move 20 lines down
* *page up*, `C-B`: move 20 lines up
* `C-D`: move 10 lines down
* `C-U`: move 10 lines up
* `G`: jump to the bottom of the buffer
* `g`: jump to the top of the buffer
* `0`: jump to the left end of the buffer
* `$`: jump to the right end of the buffer

In combination with the documentation of the other commands (see `cobib.commands`) the usage of the
TUI should be rather straight forward.

### Configuration

coBib's TUI provides a variety of configuration options.
In order to not duplicate the information here, please refer to the end of `cobib.config.example`
where all settings are listed and commented in detail.
"""

from __future__ import annotations

import curses
import fcntl
import logging
import re
import shlex
import signal
import struct
import sys
from termios import TIOCGWINSZ
from typing import IO, Any, Callable, Dict, List, Optional, Set, Tuple, Union, cast

from cobib import commands
from cobib.config import config
from cobib.utils.file_downloader import FileDownloader

from .buffer import InputBuffer, TextBuffer
from .frame import Frame
from .state import STATE, Mode

LOGGER = logging.getLogger(__name__)


class TUI:
    """coBib's curses-based TUI.

    The TUI is implemented as a class to simplify management of different windows/pads and keep a
    synchronized state most consistently.
    """

    COLOR_VALUES: Dict[str, int] = {
        "black": curses.COLOR_BLACK,
        "blue": curses.COLOR_BLUE,
        "cyan": curses.COLOR_CYAN,
        "green": curses.COLOR_GREEN,
        "magenta": curses.COLOR_MAGENTA,
        "red": curses.COLOR_RED,
        "white": curses.COLOR_WHITE,
        "yellow": curses.COLOR_YELLOW,
    }
    """A dictionary mapping color names to their `curses` values."""

    COLOR_NAMES: List[str] = [
        "top_statusbar",
        "bottom_statusbar",
        "search_label",
        "search_query",
        "cursor_line",
        "popup_help",
        "popup_stdout",
        "popup_stderr",
        "selection",
    ]
    """The list of available TUI colors."""

    ANSI_MAP: Dict[str, int] = {}
    """A dictionary mapping ANSI color codes to `curses` color pairs."""

    COMMANDS: Dict[str, Callable] = {  # type: ignore
        "Add": commands.add.AddCommand.tui,
        "Delete": commands.delete.DeleteCommand.tui,
        "Edit": commands.edit.EditCommand.tui,
        "Export": commands.export.ExportCommand.tui,
        "Filter": commands.list.ListCommand.tui_filter,
        "Help": lambda self: self.help(),
        "Modify": commands.modify.ModifyCommand.tui,
        "Open": commands.open.OpenCommand.tui,
        "Prompt": lambda self: self.execute_command(None),
        "Quit": lambda self: self.quit(),
        "Redo": commands.redo.RedoCommand.tui,
        "Search": commands.search.SearchCommand.tui,
        "Select": lambda self: self.select(),
        "Show": commands.show.ShowCommand.tui,
        "Sort": commands.list.ListCommand.tui_sort,
        "Undo": commands.undo.UndoCommand.tui,
        "Wrap": lambda self: self.viewport.wrap(),
        "x": lambda self, update: self.viewport.scroll_x(update),
        "y": lambda self, update: self.viewport.scroll_y(update),
    }
    """The dictionary of available commands."""

    HELP_DICT: Dict[str, str] = {
        "Add": "Prompts for a new entry to be added to the database.",
        "Delete": "Removes the current entry from the database.",
        "Edit": "Edits the current entry in an external EDITOR.",
        "Export": "Allows exporting the database to .bib or .zip files.",
        "Filter": "Allows filtering the list via `++/--` keywords.",
        "Help": "Displays this help.",
        "Modify": "Allows basic modification of multiple entries at once.",
        "Open": "Opens the associated file of an entry.",
        "Prompt": "Executes arbitrary coBib CLI commands in the prompt.",
        "Quit": "Closes current menu and quit's coBib.",
        "Redo": "Redoes the last undone change. Requires git-tracking!",
        "Search": "Searches the database for a given string.",
        "Select": "Adds the current entry to the interactive selection.",
        "Show": "Shows the details of an entry.",
        "Sort": "Prompts for the field to sort against (-r to reverse).",
        "Undo": "Undoes the last change. Requires git-tracking!",
        "Wrap": "Wraps the text displayed in the window.",
    }
    """The dictionary of help strings associated with their commands."""

    KEYDICT: Dict[int, Any] = {
        curses.KEY_DOWN: ("y", 1),
        curses.KEY_UP: ("y", -1),
        curses.KEY_NPAGE: ("y", 20),
        curses.KEY_PPAGE: ("y", -20),
        ord("j"): ("y", 1),
        ord("k"): ("y", -1),
        ord("g"): ("y", "g"),
        ord("G"): ("y", "G"),
        2: ("y", -20),  # CTRL-B
        4: ("y", 10),  # CTRL-D
        6: ("y", 20),  # CTRL-F
        21: ("y", -10),  # CTRL-U
        curses.KEY_LEFT: ("x", -1),
        curses.KEY_RIGHT: ("x", 1),
        ord("h"): ("x", -1),
        ord("l"): ("x", 1),
        ord("0"): ("x", 0),
        ord("$"): ("x", "$"),
    }
    """The dictionary of standard key-bindings. This will *only* be completed at runtime."""

    def __init__(self, stdscr: "curses.window", debug: bool = False) -> None:  # type: ignore
        """Initializes the curses-TUI and starts the event loop.

        Args:
            stdscr: the curses standard screen as returned by `curses.initscr`.
            debug: if True, the key-event loop is not automatically started.
        """
        LOGGER.info("Initializing TUI.")
        self.stdscr = stdscr

        # redirect stdout and stderr
        self.stdout = sys.stdout = TextBuffer()  # type: ignore
        self.stderr = sys.stderr = TextBuffer()  # type: ignore

        # register resize handler
        signal.signal(signal.SIGWINCH, self.resize_handler)

        # Clear and refresh the screen for a blank canvas
        self.stdscr.clear()
        self.stdscr.refresh()
        # Initialize layout
        curses.curs_set(0)
        self.height, self.width = self.stdscr.getmaxyx()
        LOGGER.debug("stdscr size determined to be %dx%d", self.width, self.height)
        # and colors
        LOGGER.debug("Initializing colors.")
        curses.use_default_colors()
        TUI.colors()
        # and user key mappings
        LOGGER.debug("Initializing key bindings.")
        TUI.bind_keys()

        # Initialize STATE
        LOGGER.debug("Initializing global State")
        self.STATE = STATE  # pylint: disable=invalid-name
        STATE.initialize()
        # load further configuration settings
        self.prompt_before_quit = config.tui.prompt_before_quit

        # the selection needs to be tracked outside of the State in order to persist across
        # different views
        self.selection: Set[str] = set()
        """The labels of the currently visually selected entries."""

        # Initialize top status bar
        LOGGER.debug("Populating top status bar.")
        self.topbar = curses.newwin(1, self.width, 0, 0)
        self.topbar.bkgd(" ", curses.color_pair(TUI.COLOR_NAMES.index("top_statusbar") + 1))

        # Initialize bottom status bar
        # NOTE: -2 leaves an additional empty line for the command prompt
        LOGGER.debug("Populating bottom status bar.")
        self.botbar = curses.newwin(1, self.width, self.height - 2, 0)
        self.botbar.bkgd(" ", curses.color_pair(TUI.COLOR_NAMES.index("bottom_statusbar") + 1))
        self.statusbar(self.botbar, self.infoline())

        # Initialize prompt line
        # The prompt is a pad to allow command/error prompts to exceed the terminal width.
        self.prompt = curses.newpad(1, self.width)

        # set prompt line as output source of the global file downloader
        FileDownloader().set_logger(self.prompt_print)

        # Initialize main viewport
        LOGGER.debug("Initializing viewport with Frame")
        # NOTE: -3 accounts for the top and bottom statusline as well as the command prompt
        self.viewport = Frame(self, self.height - 3, self.width)
        # populate buffer with list of reference entries
        LOGGER.debug("Populating viewport buffer.")
        self.viewport.update_list()

        if not debug:
            # start key event loop
            LOGGER.debug("Starting key event loop.")
            self.loop()
            LOGGER.info("Exiting TUI.")  # pragma: no cover

    def resize_handler(self, signum: Optional[int], frame) -> None:  # type: ignore
        # pylint: disable=unused-argument
        """Handles terminal window resizing events.

        This method gets exploited to trigger a refresh of the entire TUI window.
        In such a case it will be called as `resize_handler(None, None)`.

        Args:
            signum: signal number.
            frame: unused argument, required by the function template.
        """
        LOGGER.debug("Handling resize event.")
        if signum == signal.SIGWINCH:
            # update total dimension data
            buf = struct.pack("HHHH", 0, 0, 0, 0)
            # We use f_d = 0 as this redirects to STDIN under the hood, regardless of whether the
            # application is actually running in the foreground or in a pseudo terminal.
            buf = fcntl.ioctl(0, TIOCGWINSZ, buf)
            self.height, self.width = struct.unpack("HHHH", buf)[0:2]
        if signum is not None and not curses.is_term_resized(self.height, self.width):
            # when no signal number was given, this was a manually triggered event with the purpose
            # of completely refreshing the screen
            LOGGER.debug("Resize event did not have any effect: %dx%d", self.width, self.height)
            return
        LOGGER.debug("New stdscr dimension determined to be %dx%d", self.width, self.height)
        # actually resize the terminal
        curses.resize_term(self.height, self.width)
        # clear and refresh for a blank canvas
        self.stdscr.clear()
        self.stdscr.refresh()
        # update top statusbar
        self.topbar.resize(1, self.width)
        self.statusbar(self.topbar, STATE.topstatus)
        self.topbar.refresh()
        # update bottom statusbar
        self.botbar.resize(1, self.width)
        self.botbar.mvwin(self.height - 2, 0)
        self.statusbar(self.botbar, self.infoline())
        self.botbar.refresh()
        # update prompt
        self.prompt.resize(1, self.width)
        self.prompt.refresh(0, 0, self.height - 1, 0, self.height, self.width - 1)
        # update viewport
        self.viewport.resize(self.height - 3, self.width)

    def quit(self) -> None:
        """Breaks the key event loop or quits one TUI level.

        You can disable the final prompt before coBib quits via `config.tui.prompt_before_quit`.
        """
        if STATE.mode == Mode.LIST.value:
            LOGGER.debug("Quitting from lowest level.")
            if self.prompt_before_quit:
                msg = "Do you really want to quit coBib? [y/n] "
                curses.curs_set(1)
                self.prompt.clear()
                if len(msg) >= self.prompt.getmaxyx()[1] - 2:
                    self.prompt.resize(1, len(msg) + 2)
                self.prompt.insstr(0, 0, msg)
                self.prompt.move(0, len(msg))
                self.prompt.refresh(0, 0, self.height - 1, 0, self.height, self.width - 1)
                key = 0
                while True:
                    if key in (ord("y"), ord("Y")):
                        raise StopIteration
                    if key in (ord("n"), ord("N")):
                        LOGGER.info("User aborted quitting.")
                        curses.curs_set(0)
                        break
                    key = self.prompt.getch()
                self.prompt.clear()
                self.prompt.refresh(0, 0, self.height - 1, 0, self.height, self.width - 1)
            else:
                raise StopIteration
        else:
            LOGGER.debug("Quitting higher menu level. Falling back to list view.")
            self.viewport.revert()

    @staticmethod
    def colors() -> None:
        """Initializes the color pairs for the curses TUI."""
        # Start colors in curses
        curses.start_color()
        # parse user color configuration
        color_cfg = config.tui.colors
        colors: Dict[str, Dict[str, str]] = {col: {} for col in TUI.COLOR_NAMES}
        for attr, col in color_cfg.items():
            if attr in TUI.COLOR_VALUES:
                if not curses.can_change_color():
                    # cannot change curses default colors
                    LOGGER.warning("Curses cannot change the default colors. Skipping color setup.")
                    continue
                # update curses-internal color with HEX-color
                rgb_color = tuple(int(col.strip("#")[i : i + 2], 16) for i in (0, 2, 4))
                # curses colors range from 0 to 1000
                curses_color = tuple(col * 1000 // 255 for col in rgb_color)
                curses.init_color(TUI.COLOR_VALUES[attr], *curses_color)
            else:
                if attr[:-3] not in TUI.COLOR_NAMES:
                    LOGGER.warning("Detected unknown TUI color name specification: %s", attr[:-3])
                    continue
                colors[attr[:-3]][attr[-2:]] = col

        # initialize color pairs for TUI elements
        for idx, attr in enumerate(TUI.COLOR_NAMES):
            foreground = colors[attr].get("fg", "white")
            background = colors[attr].get("bg", "black")
            LOGGER.debug("Initiliazing color pair %d for %s", idx + 1, attr)
            curses.init_pair(idx + 1, TUI.COLOR_VALUES[foreground], TUI.COLOR_VALUES[background])
            LOGGER.debug("Adding ANSI color code for %s", attr)
            TUI.ANSI_MAP[config.get_ansi_color(attr)] = TUI.COLOR_NAMES.index(attr) + 1

    @staticmethod
    def bind_keys() -> None:
        """Binds keys according to the user configuration."""
        for command, key in config.tui.key_bindings.items():
            command = command.title()
            LOGGER.info("Binding key %s to the %s command.", key, command)
            if command not in TUI.COMMANDS.keys():
                LOGGER.warning('Unknown command "%s". Ignoring key binding.', command)
                continue
            if key == "ENTER":
                TUI.KEYDICT[10] = command  # line feed
                TUI.KEYDICT[13] = command  # carriage return
                continue
            if isinstance(key, str):
                # map key to its ASCII number
                key = ord(key)
            TUI.KEYDICT[key] = command

    @staticmethod
    def statusbar(statusline: "curses.window", text: str, attr: int = 0) -> None:  # type: ignore
        """Updates the text in the provided status bar and refreshes it.

        Args:
            statusline: single line height window used as a statusline.
            text: text to place in the statusline.
            attr: attribute number to use for the printed text.
        """
        statusline.erase()
        _, max_x = statusline.getmaxyx()
        statusline.addnstr(0, 0, text, max_x - 1, attr)
        statusline.refresh()

    @staticmethod
    def infoline() -> str:
        """Returns the available key bindings for the bottom status bar."""
        infoline = ""
        for cmd in TUI.HELP_DICT:
            # get associated key for this command
            key = config.tui.key_bindings[cmd.lower()]
            infoline += f" {key}:{cmd}"
        return infoline.strip()

    def help(self) -> None:
        # pylint: disable=consider-using-f-string
        """The Help Command.

        Opens a popup with more detailed information on the configured key bindings and short
        descriptions of the associated commands.
        """
        LOGGER.debug("Help command triggered.")
        # populate text buffer with help text
        help_text = TextBuffer()
        LOGGER.debug("Generating help text.")
        for cmd, desc in TUI.HELP_DICT.items():
            key: Union[int, str]
            for key, command in TUI.KEYDICT.items():
                if cmd == command:
                    # find mapped key
                    key = "ENTER" if key in (10, 13) else chr(key)
                    break
            # write: [key] Command: Description
            help_text.write("{:^8} {:<8} {}".format("[" + str(key) + "]", cmd + ":", desc))
        # add header section
        help_text.lines.insert(0, "{0:^{1}}".format("coBib TUI Help", help_text.width))
        help_text.lines.insert(1, "{:^8} {:<8} {}".format("Key", "Command", "Description"))
        help_text.height += 2

        # open help popup
        help_text.popup(self, background=TUI.COLOR_NAMES.index("popup_help"))

    def loop(self, debug: bool = False) -> None:
        """The key-handling event loop.

        This method takes care of reading the user's key strokes and triggering associated commands.

        Args:
            debug: if True, the key-event loop is not automatically started.
        """
        key = 0
        # key is the last character pressed
        while True:
            # handle possible keys
            try:
                if key in TUI.KEYDICT:
                    cmd = TUI.KEYDICT[key]
                    if cmd not in STATE.inactive_commands:
                        if isinstance(cmd, tuple):
                            TUI.COMMANDS[cmd[0]](self, cmd[1])
                        else:
                            TUI.COMMANDS[cmd](self)
                elif key == curses.KEY_RESIZE:
                    self.resize_handler(None, None)
            except StopIteration:
                LOGGER.debug("Stopping key event loop.")
                # raised by quit command
                sys.stdout = sys.__stdout__
                sys.stderr = sys.__stderr__
                break

            # highlight current line
            current_attributes: List[Optional[int]] = []
            for x_pos in range(0, self.viewport.pad.getmaxyx()[1]):
                x_ch = self.viewport.pad.inch(STATE.current_line, x_pos)
                # store attributes by filtering out the character text
                current_attributes.append(x_ch & curses.A_CHARTEXT)  # type: ignore
                # extract color information
                x_color = x_ch & curses.A_COLOR  # type: ignore
                color_pair_content = curses.pair_content(curses.pair_number(x_color))
                # if the current color attribute is lower in priority than the current line
                # highlighting, use that instead
                if all(col <= 0 for col in color_pair_content):
                    self.viewport.pad.chgat(
                        STATE.current_line,
                        x_pos,
                        1,
                        curses.color_pair(TUI.COLOR_NAMES.index("cursor_line") + 1),
                    )
                else:
                    # otherwise, we remove the stored attributes in order to not reset them later
                    current_attributes[-1] = None

            for stream, name in zip((self.stderr, self.stdout), ("stderr", "stdout")):
                stream.flush()
                if stream.lines:
                    stream.split()
                    LOGGER.info("sys.%s contains:\n%s", name, "\n".join(stream.lines))
                    # wrap before checking the height:
                    stream.wrap(self.width)
                    if stream.height > 1 and not debug:
                        stream.popup(self, background=TUI.COLOR_NAMES.index(f"popup_{name}"))
                    else:
                        self.prompt_print(stream.lines)
                    stream.clear()

            # Refresh the screen
            self.viewport.refresh()

            # Wait for next input
            key = self.stdscr.getch()
            LOGGER.debug("Key press registered: %s", str(key))

            # reset highlight of current line
            for x_pos in range(0, self.viewport.pad.getmaxyx()[1]):
                attr = current_attributes[x_pos]
                if attr is not None:
                    self.viewport.pad.chgat(STATE.current_line, x_pos, 1, attr)

    def select(self) -> None:
        """Toggles selection of the label currently under the cursor."""
        LOGGER.debug("Select command triggered.")
        # get current label
        label, cur_y = self.viewport.get_current_label()
        # toggle selection
        if label not in self.selection:
            LOGGER.info("Adding '%s' to the selection.", label)
            self.selection.add(label)
            # Note, that we use an additional two spaces to attempt to uniquely identify the label
            # in the list mode. Otherwise it might be possible that the same text (as used for the
            # label) can occur elsewhere in the buffer.
            # We do not need this outside of the list view because then the line indexed by `cur_y`
            # will surely only include the one label which we actually want to operate on.
            offset = "  " if STATE.mode == Mode.LIST.value else ""
            self.viewport.buffer.replace(
                cur_y,
                label + offset,
                config.get_ansi_color("selection") + label + "\x1b[0m" + offset,
            )
        else:
            LOGGER.info("Removing '%s' from the selection.", label)
            self.selection.remove(label)
            self.viewport.buffer.replace(
                cur_y,
                re.escape(config.get_ansi_color("selection")) + label + re.escape("\x1b[0m"),
                label,
            )
        # update buffer view
        self.viewport.view(ansi_map=self.ANSI_MAP)

    def prompt_print(self, text: Union[str, List[str]]) -> None:
        """Prints text to the prompt line.

        This function also handles a bug in curses which disallows 'newline'-characters [1].
        Thus, if this is the case, the first line is printed in the prompt and the total text is
        also presented in a popup. The reason for the duplicate information is too provide a little
        context on the left-over message once the popup window has been closed.

        [1] https://docs.python.org/3/library/curses.html#curses.window.addstr

        Args:
            text: the test to print to the prompt.
        """
        lines = text.strip().split("\n") if isinstance(text, str) else text
        self.prompt.clear()
        self.prompt.resize(1, max(len(lines[0]), self.width))
        self.prompt.addstr(0, 0, lines[0])
        self.prompt.refresh(0, 0, self.height - 1, 0, self.height, self.width - 1)
        if len(lines) > 1:
            buffer = TextBuffer()
            buffer.write("\n".join(lines))
            buffer.popup(self, background=TUI.COLOR_NAMES.index("popup_stderr"))

    def prompt_handler(self, command: Optional[str], symbol: str = ":") -> str:
        """Handles prompt input.

        This method starts another loop during which user key strokes are being caught and
        interpreted.

        Args:
            command: the command string to populate the prompt with.
            symbol: the prompt symbol.

        Returns:
            The final user command.
        """
        LOGGER.debug("Handling input by the user in prompt line.")
        # make cursor visible
        curses.curs_set(1)

        # populate prompt line and place cursor
        prompt_line = symbol if command is None else f"{symbol}{command} "
        self.prompt.clear()
        self.prompt.resize(1, max(len(prompt_line) + 2, self.width))
        self.prompt.addstr(0, 0, prompt_line)
        self.prompt.move(0, len(prompt_line))
        self.prompt.refresh(0, 0, self.height - 1, 0, self.height, self.width - 1)

        key = 0
        command = ""
        while True:
            # get current cursor position
            _, cur_x = self.prompt.getyx()
            # get next key
            self.prompt.nodelay(False)
            key = self.prompt.getch()
            LOGGER.debug("Key press registered: %s", str(key))
            # handle keys
            if key == 27:  # ESC
                self.prompt.nodelay(True)
                # check if it was an arrow escape sequence
                self.prompt.getch()
                arrow = self.prompt.getch()
                if arrow == -1:
                    LOGGER.debug("Prompt input aborted by the user.")
                    # if not, ESC ends the prompt
                    break
                if arrow == 68:  # left arrow key
                    LOGGER.debug("Move cursor left from arrow key input.")
                    self.prompt.move(_, cur_x - 1)
                elif arrow == 67:  # right arrow key
                    LOGGER.debug("Move cursor right from arrow key input.")
                    self.prompt.move(_, cur_x + 1)
            elif key in (8, 127):  # BACKSPACE
                if cur_x > 1:
                    LOGGER.debug("Delete the previous character in the prompt.")
                    self.prompt.delch(_, cur_x - 1)
                else:
                    LOGGER.debug("Prompt input aborted by the user.")
                    command = ""
                    break
            elif key in (10, 13):  # ENTER
                LOGGER.debug("Execute the command as prompted.")
                command = cast(bytes, self.prompt.instr(0, 1)).decode("utf-8").strip()
                break
            elif key == -1:  # no input
                break  # pragma: no cover
            else:
                # any normal key is simply echoed
                self.prompt.resize(1, max(cur_x + 2, self.width))
                self.prompt.addstr(_, cur_x, chr(key))
                self.prompt.move(_, cur_x + 1)
            self.prompt.refresh(
                0, max(0, cur_x - self.width + 2), self.height - 1, 0, self.height, self.width - 1
            )
        # leave echo mode and make cursor invisible
        curses.curs_set(0)

        # clear prompt line
        self.prompt.clear()
        self.prompt.refresh(0, 0, self.height - 1, 0, self.height, self.width - 1)

        return command

    def execute_command(
        self,
        command: Union[str, List[str]],
        out: Optional[IO[Any]] = None,
        pass_selection: bool = False,
        skip_prompt: bool = False,
    ) -> Tuple[List[str], Any]:
        """Executes a command.

        Args:
            command: the command to execute.
            out: the output stream to redirect stdout to.
            pass_selection: whether to the pass the current TUI selection in the executed command
                arguments.
            skip_prompt: whether to skip the user prompt for further commands.

        Returns:
            A pair with the first element being the list with the executed command to allow further
            handling and the second element being whatever is returned by the command.
        """
        if not skip_prompt:
            command = self.prompt_handler(cast(str, command))
            # split command into separate arguments for cobib
            try:
                command = shlex.split(command)
            except ValueError:
                LOGGER.error("Invalid command: %s", command)
                return ([""], None)

        command = cast(List[str], command)

        # process command if it is non empty and actually has arguments
        result = None
        if command and command[0]:
            LOGGER.debug("Processing the command: %s", " ".join(command))
            # run command
            sys.stdin = InputBuffer(buffer=self.stdout, tui=self)  # type: ignore
            subcmd = getattr(commands, command[0].title() + "Command")()
            try:
                if pass_selection:
                    command += ["--"]
                    command.extend(list(self.selection))
                result = subcmd.execute(command[1:], out=out)
            except SystemExit:  # pragma: no cover
                pass  # pragma: no cover
            sys.stdin = sys.__stdin__
        # return command to enable additional handling by function caller
        return (command, result)



r"""coBib's List command.

This command simply lists the entries in the database:
```
cobib list
```

### Basic Options

It provides some very basic output manipulation options.

You can disable the shortening of the output:
```
cobib list --long
```
You can reverse the output:
```
cobib list --reverse
```
You can sort the list according to a database field:
```
cobib list --sort year
```
(In the TUI, this is available via the `s` key, by default.)


### Filters

However, the arguably most useful feature of this command are the *filters*.
These are a set of keyword arguments which are registered at runtime depending on the fields which
appear in your database.
To give an example, you can find the following filters in the output of `cobib list --help`:
```
++author AUTHOR
--author AUTHOR
++year YEAR
--year YEAR
```
As you can see, each field is used twice: once with a `++` and once with a `--` prefix.
These allow you to specify positive and negative matches, respectively.
For example:
```
cobib list ++year 2020
```
will list *only* entries whose `year` contains `2020`.
On the contrary:
```
cobib list --year 2020
```
will print all those entries whose `year` does *not* contain `2020`.

You can combine multiple filters to narrow your selection down further.
For example:
```
cobib list ++year 2020 ++author Rossmannek
```
will list entries whose `year` contains `2020` *and* whose `author` field contains `Rossmannek`.

Note, that you by default you can use the `f` key in the TUI to add filters to displayed list of
entries.

There are some aspects to take note of here:
1. By default, multiple filters are combined with logical `AND`s. You can specify `--or` to
   overwrite this to logical `OR`s. This will also apply to all filters of the specified command.
2. All entries are treated as `str`. Thus, `++year 20` will match anything *containing* `20`.

As of version v3.2.0, the filter arguments are evaluated as regex patterns allowing you to do things
like the following:
```
cobib list ++label "\D+_\d+"
```
This will list all entries whose labels are formatted as `"<non-digit characters>_<digits>"`.
"""

from __future__ import annotations

import argparse
import logging
import textwrap
from collections import defaultdict
from typing import IO, TYPE_CHECKING, Any, Dict, Generator, List, Optional, Set, Tuple

from cobib.config import Event
from cobib.database import Database

from .base_command import ArgumentParser, Command

LOGGER = logging.getLogger(__name__)

if TYPE_CHECKING:
    import cobib.tui


class ListCommand(Command):
    """The List Command."""

    name = "list"

    # pylint: disable=too-many-branches
    def execute(self, args: List[str], out: Optional[IO[Any]] = None) -> Optional[List[str]]:
        """Lists the entries in the database.

        This command simply lists the labels and titles of the entries in the database.
        By default, it lists them in the order in which they appear in the database.
        However, the order as well as selection of entries to be listed can be configured through
        additional arguments.
        Please refer to the [Filters](#filters) section above for more information.
        Note, that the filtering mechanisms leverages `cobib.database.Entry.matches` to do the
        actual filter-matching.

        Args:
            args: a sequence of additional arguments used for the execution. The following values
                are allowed for this command:
                    * `-l`, `--long`: if specified, the columns of the output table will be not be
                      shortened.
                    * `-r`, `--reverse`: if specified, the entries will be listed in reverse order.
                      This is especially useful in the TUI (where it is enabled by default) because
                      it puts the last added entries at the top of the window. When using the
                      command-line interface it is disabled by default, because this puts the last
                      added entries at the bottom, just above the new command-line prompt.
                    * `-x`, `--or`: if specified, multiple filters will be combined with logical OR
                      rather than the default logical AND.
                    * `-s`, `--sort`: you can specify an arbitrary `cobib.database.Entry.data` field
                      name which should be used for sorting the listed entries. This will
                      automatically include a column for this field in the output table.
                    * in addition to the options above, [Filter keyword arguments](#filters) are
                      registered at runtime based on the fields available in the database. Please
                      refer the that section for more information.
            out: the output IO stream. This defaults to `None`.

        Returns:
            A list with the filtered and sorted labels.
        """
        LOGGER.debug("Starting List command.")
        if "--" in args:
            args.remove("--")  # pragma: no cover
        parser = ArgumentParser(
            prog="list", description="List subcommand parser.", prefix_chars="+-"
        )
        parser.add_argument(
            "-l",
            "--long",
            action="store_true",
            help="print table in long format (i.e. don't shorten lines)",
        )
        parser.add_argument(
            "-r", "--reverse", action="store_true", help="reverses the listing order"
        )
        parser.add_argument("-s", "--sort", help="specify column along which to sort the list")
        parser.add_argument(
            "-x",
            "--or",
            dest="OR",
            action="store_true",
            help="concatenate filters with OR instead of AND",
        )
        unique_keys: Set[str] = {"label"}
        LOGGER.debug("Gathering possible filter arguments.")
        for entry in Database().values():
            unique_keys.update(entry.data.keys())
        for key in sorted(unique_keys):
            parser.add_argument(
                "++" + key, type=str, action="append", help="include elements with matching " + key
            )
            parser.add_argument(
                "--" + key, type=str, action="append", help="exclude elements with matching " + key
            )

        try:
            largs = parser.parse_args(args)
        except argparse.ArgumentError as exc:
            LOGGER.error(exc.message)
            return None

        Event.PreListCommand.fire(largs)

        LOGGER.debug("Constructing filter.")
        _filter: Dict[Tuple[str, bool], List[Any]] = defaultdict(list)
        for key, val in largs.__dict__.items():
            if key in ["OR", "long", "sort", "reverse"] or val is None:
                continue
            if not isinstance(val, list):
                val = [val]
            for i in val:
                for idx, obj in enumerate(args):
                    if i == obj:
                        index: Tuple[str, bool] = (key, args[idx - 1][0] == "+")
                        _filter[index].append(i)
                        break
        LOGGER.debug("Final filter configuration: %s", dict(_filter))
        if largs.OR:
            LOGGER.debug("Filters are combined with logical ORs!")

        columns = ["label", "title"]
        if largs.sort and largs.sort not in columns:
            # insert columns which are sorted by at front of list view
            LOGGER.debug('Sorting by "%s".', largs.sort)
            columns.insert(1, largs.sort)
        # filtered columns are still appended
        columns.extend([arg[0] for arg in _filter.keys() if arg[0] not in columns])
        widths = [0] * len(columns)
        labels = []
        table = []
        for key, entry in Database().items():
            if entry.matches(_filter, largs.OR):
                LOGGER.debug('Entry "%s" matches the filter.', key)
                labels.append(key)
                table.append([entry.stringify().get(c, "") for c in columns])
                if largs.long:
                    table[-1][1] = table[-1][1]
                else:
                    table[-1][1] = textwrap.shorten(table[-1][1], 80, placeholder="...")
                widths = [max(widths[col], len(table[-1][col])) for col in range(len(widths))]
        LOGGER.debug("Column widths determined to be: %s", widths)
        if largs.sort:
            LOGGER.debug("Sorting table in %s order.", "reverse" if largs.reverse else "normal")
            labels, table = zip(  # type: ignore
                *sorted(
                    zip(labels, table),
                    reverse=largs.reverse,
                    key=lambda row: row[1][columns.index(largs.sort)],  # type: ignore
                )
            )
        elif largs.reverse:
            # do not sort, but reverse
            LOGGER.debug("Reversing order.")
            labels, table = labels[::-1], table[::-1]
        for row in table:
            print("  ".join([f"{col: <{wid}}" for col, wid in zip(row, widths)]), file=out)

        Event.PostListCommand.fire(labels)

        return list(labels)

    @staticmethod
    def tui_sort(tui: cobib.tui.TUI) -> None:
        """TUI command interface.

        This function serves as the entry-point from the `cobib.tui.tui.TUI` instance for the
        **sort** functionality provided by the `ListCommand`.

        Args:
            tui: the runtime-instance of coBib's TUI.
        """
        try:
            # first, remove any previously used sort argument
            sort_arg_idx = tui.STATE.list_args.index("-s")
            prev = tui.STATE.list_args.pop(sort_arg_idx + 1)
            LOGGER.debug('Removing previous sort argument: "%s"', prev)
            tui.STATE.list_args.pop(sort_arg_idx)
        except ValueError:
            pass

        # add the sort option to the arguments
        tui.STATE.list_args += ["-s"]

        done = False
        # run the actual command
        for command in ListCommand.tui(tui):
            if done:
                raise RuntimeError("Received multiple commands to be processed. Aborting!")

            try:
                sort_arg_idx = command.index("-s")
                if sort_arg_idx + 1 >= len(command):
                    # sort argument got removed
                    raise ValueError
                tui.STATE.list_args += [command[sort_arg_idx + 1]]
                LOGGER.debug('Using sort argument: "%s"', tui.STATE.list_args[-1])
            except ValueError:
                tui.STATE.list_args.remove("-s")

            # ensure we never process more than one yield statement
            done = True

    @staticmethod
    def tui_filter(tui: cobib.tui.TUI) -> None:
        """TUI command interface.

        This function serves as the entry-point from the `cobib.tui.tui.TUI` instance for the
        **filter** functionality provided by the `ListCommand`.

        Args:
            tui: the runtime-instance of coBib's TUI.
        """
        done = False
        for command in ListCommand.tui(tui):
            if done:
                raise RuntimeError("Received multiple commands to be processed. Aborting!")

            # first, pop all filters from tui.STATE.list_args
            indices_to_pop = []
            # enumerate words in current list arguments
            prev_args = list(enumerate(tui.STATE.list_args))
            # iterate in reverse to ensure popping indices remain correct after popping a few
            prev_args.reverse()
            for idx, p_arg in prev_args:
                if p_arg[:2] in ("++", "--"):
                    # matches a filter: current index is type and one larger is the key
                    LOGGER.debug(prev_args)
                    LOGGER.debug(
                        'Removing filter from prompt: "%s"',
                        " ".join(tui.STATE.list_args[idx : idx + 2]),
                    )
                    indices_to_pop.extend([idx + 1, idx])
            for idx in indices_to_pop:
                tui.STATE.list_args.pop(idx)
            # then, add all new filter (type, key) pairs
            for idx, n_arg in enumerate(command):
                if n_arg[:2] in ("++", "--"):
                    LOGGER.debug('Adding filter to prompt: "%s"', " ".join(command[idx : idx + 2]))
                    tui.STATE.list_args.extend(command[idx : idx + 2])
            # reset current line position to top
            tui.STATE.current_line = 0

            # ensure we never process more than one yield statement
            done = True

    @staticmethod
    def tui(tui: cobib.tui.TUI) -> Generator[List[str], None, None]:
        # pdoc will inherit the docstring from the base class
        # noqa: D102
        LOGGER.debug("List command triggered from TUI.")
        LOGGER.debug("Clearing current buffer contents.")
        tui.viewport.clear()
        # handle input via prompt
        command, _ = tui.execute_command(
            "list " + " ".join(tui.STATE.list_args), out=tui.viewport.buffer  # type: ignore
        )
        # after the command has been executed in the prompt handler, the `command` variable will
        # contain the contents of the prompt
        LOGGER.debug("Post-process ListCommand arguments for consistent prompt.")
        if command:
            # always ensure the keyword arguments are consistent
            for arg, name in [("-r", "reverse"), ("-x", "OR")]:
                if arg in command and arg not in tui.STATE.list_args:
                    LOGGER.debug('Adding "%s" list argument.', name)
                    tui.STATE.list_args.insert(1, arg)
                elif arg not in command and arg in tui.STATE.list_args:
                    LOGGER.debug('Removing "%s" list argument.', name)
                    tui.STATE.list_args.remove(arg)

            # hand command to callable frame for potential further processing
            yield command

        # populate buffer with the list
        LOGGER.debug("Populating buffer with ListCommand results.")
        tui.STATE.mode = "list"
        tui.STATE.previous_line = -1
        tui.STATE.inactive_commands = []
        tui.viewport.view()
        # update database list
        tui.viewport.update_list()



"""An example configuration for coBib.

Since version 3.0 coBib is configured through a Python file.
For documentation purposes this example configuration lists all possible settings with their default
values and detailed explanations.

Internally, coBib's configuration is nothing but a (nested) Python dictionary. However, for ease of
usage all of its fields are also exposed as attributes. This means, that the following two lines are
equivalent:

```python
config['database']['git'] = True
config.database.git = True
```
"""

# Generally, you won't need these, but the default configuration relies on them.
import os
import sys

# To get started you must import coBib's configuration.
from cobib.config import LabelSuffix, config

# Now, you are all set to apply your own settings.


# LOGGING
# You can specify the default logfile location.
config.logging.logfile = "~/.cache/cobib/cobib.log"
# You can also set the location for the cached version number based on which coBib shows you the
# latest changes. You can set this to `None` to disable this functionality entirely.
config.logging.version = "~/.cache/cobib/version"

# COMMANDS
# These settings affect some command specific behavior.

# You can specify the default bibtex entry type.
config.commands.edit.default_entry_type = "article"

# You can specify the editor program.
config.commands.edit.editor = os.environ.get("EDITOR", "vim")
# Note, that this default will respect your `$EDITOR` environment setting and fallback to `vim` if
# that variable is not set.

# You can specify a custom command which will be used to `open` files associated with your entries.
config.commands.open.command = "xdg-open" if sys.platform.lower() == "linux" else "open"

# You can specify a custom grep tool which will be used to search through your database and any
# associated files. The default tool (`grep`) will not provide results for attached PDFs but other
# tools such as [ripgrep-all](https://github.com/phiresky/ripgrep-all) will.
config.commands.search.grep = "grep"
# If you want to specify additional arguments for your grep command, you can specify them as a list
# of strings in the following setting. Note, that GNU's grep understands extended regex patterns
# even without specifying `-E`.
config.commands.search.grep_args = []

# You can specify whether searches should be performed case-insensitive.
config.commands.search.ignore_case = False


# DATABASE
# These settings affect the database in general.

# You can specify the path to the database YAML file. You can use a `~` to represent your `$HOME`
# directory.
config.database.file = "~/.local/share/cobib/literature.yaml"

# coBib can integrate with `git` in order to automatically track the history of your database.
# However, by default, this option is disabled. If you want to enable it, simply change the
# following setting to `True` and initialize your database with `cobib init --git`.
# Warning: Before enabling this setting you must ensure that you have set up git properly by setting
# your name and email address.
config.database.git = False

# DATABASE.FORMAT
# You can also specify some aspects about the format of the database.

# You can specify a default label format which will be used for the database entry keys. The format
# of this option follows the f-string like formatting of modifications (see also the documentation
# of the [ModifyCommand](https://mrossinek.gitlab.io/cobib/cobib/commands/modify.html)). The default
# configuration value leaves the label unchanged compared to the metadata provided by the source
# from which the entry gets added. A more useful example is `"{author.split()[1]}{year}"` which
# takes the surname of the first author and immediately appends the publication year.
config.database.format.label_default = "{label}"

# You can specify the suffix format which is used to disambiguate labels if a conflict would occur.
# This option takes a tuple of length 2, where the first entry is the string separating the proposed
# label from the enumerator and the second one is one of the enumerators provided in the
# `config.LabelSuffix` object. The available enumerators are:
#   - ALPHA: a, b, ...
#   - CAPTIAL: A, B, ...
#   - NUMERIC: 1, 2, ...
config.database.format.label_suffix = ("_", LabelSuffix.ALPHA)

# You can specify whether latex warnings should not be ignored during the escaping of special
# characters. This is a simple option which gets passed on to the internally used `pylatexenc`
# library.
config.database.format.suppress_latex_warnings = True

# DATABASE.STRINGIFY
# You can customize the functions which convert non-string fields.

# Three fields are currently explicitly stored as lists internally. Upon conversion to the BibTeX
# format, these need to be converted to a basic string. In this process the entries of the list will
# be joined using the separators configured by the following settings.
config.database.stringify.list_separator.file = ", "
config.database.stringify.list_separator.tags = ", "
config.database.stringify.list_separator.url = ", "

# PARSERS
# These settings affect some parser specific behavior.

# You can specify whether the bibtex-parser should ignore non-standard bibtex entry types.
config.parsers.bibtex.ignore_non_standard_types = False


# TUI
# These settings affect the functionality and look of the TUI.

# You can specify a list of default arguments for the default list view.
config.tui.default_list_args = ["-l"]

# You can disable the prompt before quitting coBib by turning off the following setting:
config.tui.prompt_before_quit = True

# You can specify whether the list view of the TUI should be reversed. By default, this is enabled,
# because this will place the most recently added entries at the top of the TUI.
config.tui.reverse_order = True

# You can specify a scroll offset. This corresponds to the number of lines which will be kept
# visible above or below the cursor line while scrolling the buffer. If you set this number to
# something very large (e.g. 99) you can pin the cursor line to the center of the window during
# scrolling.
# Note: if you are a Vim user, this setting will feel similar to `scrolloff`.
config.tui.scroll_offset = 3

# TUI.COLORS
# With the following color settings you can change the look of the TUI. Each of these settings
# accepts any of the following color names: `black`, `red`, `green`, `yellow`, `blue`, `magenta`,
# `cyan` and `white`.
config.tui.colors.cursor_line_fg = "white"
config.tui.colors.cursor_line_bg = "cyan"
config.tui.colors.top_statusbar_fg = "black"
config.tui.colors.top_statusbar_bg = "yellow"
config.tui.colors.bottom_statusbar_fg = "black"
config.tui.colors.bottom_statusbar_bg = "yellow"
config.tui.colors.search_label_fg = "blue"
config.tui.colors.search_label_bg = "black"
config.tui.colors.search_query_fg = "red"
config.tui.colors.search_query_bg = "black"
config.tui.colors.popup_help_fg = "white"
config.tui.colors.popup_help_bg = "green"
config.tui.colors.popup_stdout_fg = "white"
config.tui.colors.popup_stdout_bg = "blue"
config.tui.colors.popup_stderr_fg = "white"
config.tui.colors.popup_stderr_bg = "red"
config.tui.colors.selection_fg = "white"
config.tui.colors.selection_bg = "magenta"

# Note: if your terminal supports it, you can even try to override the color specifications right
# from within coBib. The check whether your terminal supports this relies on the
# `curses.can_change_color()` function, which is more or less documented
# [here](https://docs.python.org/3/library/curses.html#curses.can_change_color).
# You can attempt to get this to work by overwriting the named colors with a `#RRGGBB` value like
# so:
#     config.tui.colors.black = `#222222`
# , which changes the black definition into a more mallow gray.

# TUI.KEY_BINDINGS
# You can also change the default key bindings of the TUI by overwriting any of the following
# settings with a different key.
config.tui.key_bindings.add = "a"
config.tui.key_bindings.delete = "d"
config.tui.key_bindings.edit = "e"
config.tui.key_bindings.export = "x"
config.tui.key_bindings.filter = "f"
config.tui.key_bindings.help = "?"
config.tui.key_bindings.modify = "m"
config.tui.key_bindings.open = "o"
config.tui.key_bindings.prompt = ":"
config.tui.key_bindings.quit = "q"
config.tui.key_bindings.redo = "r"
config.tui.key_bindings.search = "/"
config.tui.key_bindings.select = "v"
config.tui.key_bindings.show = "ENTER"
config.tui.key_bindings.sort = "s"
config.tui.key_bindings.undo = "u"
config.tui.key_bindings.wrap = "w"
# Note, the exception of the key for `show` which is set to the custom `ENTER` string. When coBib
# encounters this string it will automatically map to the ASCII codes 10 and 13 (corresponding to
# the `line feed` and `carriage return`, respectively). Any other string is interpreted a single
# character whose ASCII value is used as the trigger.

# UTILS

# You can specify the default download location for associated files.
config.utils.file_downloader.default_location = "~/.local/share/cobib"

# You can provide rules to map from a journal's landing page URL to its PDF URL. To do so, you must
# insert an entry into the following dictionary, with a regex-pattern matching the journal's landing
# page URL and a value being the PDF URL. E.g.:
#
#     config.utils.file_downloader.url_map[
#         r"(.+)://aip.scitation.org/doi/([^/]+)"
#     ] = r"\1://aip.scitation.org/doi/pdf/\2"
#
#     config.utils.file_downloader.url_map[
#         r"(.+)://quantum-journal.org/papers/([^/]+)"
#     ] = r"\1://quantum-journal.org/papers/\2/pdf/"
#
# Make sure to use raw Python strings to ensure proper backslash-escaping.
config.utils.file_downloader.url_map = {}

# You can specify a list of journal abbreviations. This list should be formatted as tuples of the
# form: `(full journal name, abbreviation)`. The abbreviation should include any necessary
# punctuation which can be excluded upon export (see also `cobib export --help`).
config.utils.journal_abbreviations = []

# EVENTS
# coBib allows you to register hooks on various events to further customize its behavior to your
# liking. Although these functions will be registered in the following dictionary, we recommend you
# to use the function-decorators as explained below.
config.events = {}
# To subscribe to a certain event do something similar to the following:
#
#     from os import system
#     from pathlib import Path
#     from cobib.config import Event
#
#     @Event.PostInitCommand.subscribe
#     def add_remote(root: Path, file: Path) -> None:
#         system(f"git -C {root} remote add origin https://github.com/user/repo")
#
# Note, that the typing is required for the config validation to pass!
# For more information refer to the
# [online documentation](https://mrossinek.gitlab.io/cobib/cobib/config/event.html).



"""Tests for coBib's ListCommand."""
# pylint: disable=no-self-use,unused-argument

from __future__ import annotations

import contextlib
import os
from argparse import Namespace
from io import StringIO
from itertools import zip_longest
from shutil import copyfile
from typing import TYPE_CHECKING, Any, List, Type

import pytest

from cobib.commands import ListCommand
from cobib.config import Event, config
from cobib.database import Database

from .. import get_resource
from ..tui.tui_test import TUITest
from .command_test import CommandTest

if TYPE_CHECKING:
    import cobib.commands


class TestListCommand(CommandTest, TUITest):
    """Tests for coBib's ListCommand."""

    def get_command(self) -> Type[cobib.commands.base_command.Command]:
        # noqa: D102
        return ListCommand

    @pytest.mark.parametrize(
        ["args", "expected"],
        [
            [[], ["einstein", "latexcompanion", "knuthwebsite"]],
            [["-r"], ["knuthwebsite", "latexcompanion", "einstein"]],
            [["-s", "year"], ["knuthwebsite", "einstein", "latexcompanion"]],
            [["-r", "-s", "year"], ["latexcompanion", "einstein", "knuthwebsite"]],
            [["++author", "Einstein"], ["einstein"]],
            [["--author", "Einstein"], ["latexcompanion", "knuthwebsite"]],
            [["++author", "Einstein", "++author", "Knuth"], []],
            [["-x", "++author", "Einstein", "++author", "Knuth"], ["einstein", "knuthwebsite"]],
        ],
    )
    def test_command(self, setup: Any, args: List[str], expected: List[str]) -> None:
        """Test the command itself.

        Args:
            setup: the `tests.commands.command_test.CommandTest.setup` fixture.
            args: the arguments to pass to the command.
            expected: the expected list of labels.
        """
        # redirect output of list to string
        file = StringIO()
        tags = ListCommand().execute(args, out=file)
        assert tags == expected
        # TODO: actually do more assertion on the table output (awaiting ListCommand refactoring)
        for line in file.getvalue().split("\n"):
            if line.startswith("label") or all(c in "- " for c in line):
                # skip table header
                continue
            assert line.split()[0] in expected

    def test_missing_keys(self, setup: Any) -> None:
        """Asserts issue #1 is fixed.

        When a key is queried which is not present in all entries, the list command should return
        normally.

        Args:
            setup: the `tests.commands.command_test.CommandTest.setup` fixture.
        """
        # redirect output of list to string
        file = StringIO()
        tags = ListCommand().execute(["++year", "1905"], out=file)
        expected = ["einstein"]
        assert tags == expected
        for line in file.getvalue().split("\n"):
            if line.startswith("label") or all(c in "- " for c in line):
                # skip table header
                continue
            assert line.split()[0] in expected

    @pytest.mark.parametrize(
        ["expected"],
        [
            [["einstein", "latexcompanion", "knuthwebsite"]],
        ],
    )
    # other variants are already covered by test_command
    def test_cmdline(
        self,
        setup: Any,
        monkeypatch: pytest.MonkeyPatch,
        capsys: pytest.CaptureFixture[str],
        expected: List[str],
    ) -> None:
        """Test the command-line access of the command.

        Args:
            setup: the `tests.commands.command_test.CommandTest.setup` fixture.
            monkeypatch: the built-in pytest fixture.
            capsys: the built-in pytest fixture.
            expected: the expected list of labels.
        """
        self.run_module(monkeypatch, "main", ["cobib", "list"])
        for line, truth in zip_longest(capsys.readouterr().out.strip().split("\n"), expected):
            # we wrap the list into an iterator in order to handle an empty list, too
            assert next(iter(line.split()), None) == truth

    def test_tui(self, setup: Any) -> None:
        """Test the TUI access of the command.

        Args:
            setup: the `tests.commands.command_test.CommandTest.setup` fixture.
        """

        def assertion(screen, logs, **kwargs):  # type: ignore
            expected_screen = [
                r"knuthwebsite    Knuth: Computers and Typesetting",
                r"latexcompanion  The \LaTeX\ Companion",
                r"einstein        Zur Elektrodynamik bewegter K{\"o}rper",
            ]
            for line, truth in zip(expected_screen, screen.display[1:]):
                assert line == truth.strip()
            expected_log = [
                ("cobib.commands.list", 10, "Starting List command."),
                ("cobib.commands.list", 10, "Gathering possible filter arguments."),
                ("cobib.commands.list", 10, "Constructing filter."),
                ("cobib.commands.list", 10, "Final filter configuration: {}"),
                ("cobib.commands.list", 10, 'Entry "einstein" matches the filter.'),
                ("cobib.commands.list", 10, 'Entry "latexcompanion" matches the filter.'),
                ("cobib.commands.list", 10, 'Entry "knuthwebsite" matches the filter.'),
                ("cobib.commands.list", 10, "Column widths determined to be: [14, 38]"),
                ("cobib.commands.list", 10, "Reversing order."),
            ]
            assert [log for log in logs if log[0] == "cobib.commands.list"] == expected_log

        self.run_tui("", assertion, {})

    @pytest.mark.parametrize(
        ["keys"],
        [
            ["sauthor\n"],
            ["syear\nsauthor\n"],
        ],
    )
    def test_tui_sort(self, setup: Any, keys: str) -> None:
        """Test the TUI `sort` command.

        Args:
            setup: the `tests.commands.command_test.CommandTest.setup` fixture.
            keys: the string of keys to pass to the TUI.
        """

        def assertion(screen, logs, **kwargs):  # type: ignore
            expected_screen = [
                "latexcompanion  Michel Goossens and Frank Mittelbach and Alexander Samarin  The",
                "knuthwebsite    Donald Knuth                                                Knut",
                "einstein        Albert Einstein                                             Zur",
            ]
            for line, truth in zip(expected_screen, screen.display[1:]):
                assert line == truth.strip()
            expected_log = [
                ("cobib.commands.list", 10, "List command triggered from TUI."),
                ("cobib.commands.list", 10, "Clearing current buffer contents."),
                ("cobib.commands.list", 10, "Starting List command."),
                ("cobib.commands.list", 10, "Gathering possible filter arguments."),
                ("cobib.commands.list", 10, "Constructing filter."),
                ("cobib.commands.list", 10, "Final filter configuration: {}"),
                ("cobib.commands.list", 10, 'Sorting by "author".'),
                ("cobib.commands.list", 10, 'Entry "einstein" matches the filter.'),
                ("cobib.commands.list", 10, 'Entry "latexcompanion" matches the filter.'),
                ("cobib.commands.list", 10, 'Entry "knuthwebsite" matches the filter.'),
                ("cobib.commands.list", 10, "Column widths determined to be: [14, 58, 38]"),
                ("cobib.commands.list", 10, "Sorting table in reverse order."),
                (
                    "cobib.commands.list",
                    10,
                    "Post-process ListCommand arguments for consistent prompt.",
                ),
                ("cobib.commands.list", 10, 'Using sort argument: "author"'),
                ("cobib.commands.list", 10, "Populating buffer with ListCommand results."),
            ]
            true_log = [log for log in logs if log[0] == "cobib.commands.list"]
            try:
                for log, truth in zip(true_log[true_log.index(expected_log[0]) :], expected_log):
                    assert log == truth
            except AssertionError:
                # in the second parametrized test we simply check that the previous sort label was
                # removed correctly
                for msg in [
                    'Sorting by "year".',
                    'Removing previous sort argument: "year"',
                    'Sorting by "author".',
                ]:
                    assert ("cobib.commands.list", 10, msg) in true_log

        self.run_tui(keys, assertion, {})

    @pytest.mark.parametrize(
        ["keys"],
        [
            ["f++label einstein\n"],
            ["f++label knuthwebsite\nf\b\b\b\b\b\b\b\b\b\b\b\b\beinstein\n"],
        ],
    )
    def test_tui_filter(self, setup: Any, keys: str) -> None:
        """Test the TUI `filter` command.

        Args:
            setup: the `tests.commands.command_test.CommandTest.setup` fixture.
            keys: the string of keys to pass to the TUI.
        """

        def assertion(screen, logs, **kwargs):  # type: ignore
            expected_screen = [
                r"einstein  Zur Elektrodynamik bewegter K{\"o}rper",
            ]
            for line, truth in zip(expected_screen, screen.display[1:]):
                assert line == truth.strip()
            expected_log = [
                ("cobib.commands.list", 10, "List command triggered from TUI."),
                ("cobib.commands.list", 10, "Clearing current buffer contents."),
                ("cobib.commands.list", 10, "Starting List command."),
                ("cobib.commands.list", 10, "Gathering possible filter arguments."),
                ("cobib.commands.list", 10, "Constructing filter."),
                (
                    "cobib.commands.list",
                    10,
                    "Final filter configuration: {('label', True): ['einstein']}",
                ),
                ("cobib.commands.list", 10, 'Entry "einstein" matches the filter.'),
                ("cobib.commands.list", 10, "Column widths determined to be: [8, 38]"),
                ("cobib.commands.list", 10, "Reversing order."),
                (
                    "cobib.commands.list",
                    10,
                    "Post-process ListCommand arguments for consistent prompt.",
                ),
                ("cobib.commands.list", 10, 'Adding filter to prompt: "++label einstein"'),
                ("cobib.commands.list", 10, "Populating buffer with ListCommand results."),
            ]
            true_log = [log for log in logs if log[0] == "cobib.commands.list"]
            try:
                for log, truth in zip(true_log[true_log.index(expected_log[0]) :], expected_log):
                    assert log == truth
            except AssertionError:
                # in the second parametrized test we simply check that the previous filters were
                # removed correctly
                for msg in [
                    "Final filter configuration: {('label', True): ['knuthwebsite']}",
                    'Adding filter to prompt: "++label knuthwebsite"',
                    'Removing filter from prompt: "++label knuthwebsite"',
                    "Final filter configuration: {('label', True): ['einstein']}",
                    'Adding filter to prompt: "++label einstein"',
                ]:
                    assert ("cobib.commands.list", 10, msg) in true_log

        self.run_tui(keys, assertion, {})

    @pytest.mark.parametrize(
        ["keys", "messages"],
        [
            ["s\b\b\b\b\b\b\n", ['Removing "reverse" list argument.']],
            ["s\b\b\b\b\b\b\ns\b\b\b-r\n", ['Adding "reverse" list argument.']],
            ["f++label einstein -x\n", ['Adding "OR" list argument.']],
            ["f-x\nf\b\b\b\b\b\b\b\n", ['Removing "OR" list argument.']],
        ],
    )
    def test_tui_argument_unification(self, setup: Any, keys: str, messages: List[str]) -> None:
        """Test that the TUI unifies keyword arguments.

        Args:
            setup: the `tests.commands.command_test.CommandTest.setup` fixture.
            keys: the string of keys to pass to the TUI.
            messages: the expected list of log messages.
        """

        def assertion(screen, logs, **kwargs):  # type: ignore
            true_log = [log for log in logs if log[0] == "cobib.commands.list"]
            for msg in messages:
                assert ("cobib.commands.list", 10, msg) in true_log

        self.run_tui(keys, assertion, {})

    # manually overwrite this test because we must populate the database with actual data
    def test_handle_argument_error(self, caplog: pytest.LogCaptureFixture) -> None:
        """Test handling of ArgumentError.

        Args:
            caplog: the built-in pytest fixture.
        """
        # use temporary config
        config.database.file = self.COBIB_TEST_DIR / "database.yaml"
        config.database.git = True

        # load temporary database
        os.makedirs(self.COBIB_TEST_DIR, exist_ok=True)
        copyfile(get_resource("example_literature.yaml"), config.database.file)
        Database().read()

        try:
            super().test_handle_argument_error(caplog)
        finally:
            # clean up file system
            os.remove(config.database.file)
            # clean up database
            Database().clear()
            # clean up config
            config.defaults()

    def test_event_pre_list_command(self, setup: Any) -> None:
        """Tests the PreListCommand event."""

        @Event.PreListCommand.subscribe
        def hook(largs: Namespace) -> None:
            largs.author = None

        assert Event.PreListCommand.validate()

        expected = [
            'einstein        Zur Elektrodynamik bewegter K{\\"o}rper',
            "latexcompanion  The \\LaTeX\\ Companion",
            "knuthwebsite    Knuth: Computers and Typesetting",
        ]

        with contextlib.redirect_stdout(StringIO()) as out:
            ListCommand().execute(["++author", "Einstein"])

            for truth, exp in zip(out.getvalue().split("\n"), expected):
                assert truth.strip() == exp

    def test_event_post_list_command(self, setup: Any) -> None:
        """Tests the PostListCommand event."""

        @Event.PostListCommand.subscribe
        def hook(labels: List[str]) -> None:
            print(labels)

        assert Event.PostListCommand.validate()

        with contextlib.redirect_stdout(StringIO()) as out:
            ListCommand().execute([])

            assert (
                out.getvalue().split("\n")[-2] == "['einstein', 'latexcompanion', 'knuthwebsite']"
            )



"""coBib's logging module.

This module provides utility methods to set up logging to different handlers.
"""

import logging
import logging.handlers
import sys
from pathlib import Path
from typing import Optional, Union

from pkg_resources import get_distribution

from .rel_path import RelPath


class _StderrHandler(logging.StreamHandler):
    """A logging handler hard-coded to `sys.stderr`.

    The reason for explicitly deriving this class, is that Python's `logging.StreamHandler` does not
    respect stream redirection. However, for coBib's TUI this is an important requirement which can
    be achieved by a runtime check of the stream during the `emit` method.
    """

    def emit(self, record: logging.LogRecord) -> None:
        # pdoc will inherit the docstring from the base class
        # noqa: D102
        self.stream = sys.stderr
        super().emit(record)


def get_stream_handler(level: int = logging.WARNING) -> logging.StreamHandler:
    """Returns a basic StreamHandler logging to `sys.stderr`.

    Args:
        level: the logging level of this handler.
    """
    formatter = logging.Formatter(fmt="[%(levelname)s] %(message)s")

    handler = _StderrHandler()
    handler.setLevel(level)
    handler.setFormatter(formatter)

    return handler


def get_file_handler(
    level: Union[str, int] = "INFO", logfile: Optional[Union[str, Path]] = None
) -> logging.handlers.RotatingFileHandler:
    """Returns a `RotatingFileHandler`.

    Args:
        level: the handler's logging level.
        logfile: the output path for the log file.
    """
    if logfile is None:
        # pylint: disable=import-outside-toplevel
        from cobib.config import config

        logfile = config.logging.logfile

    path = RelPath(logfile).path
    path.parent.mkdir(parents=True, exist_ok=True)

    formatter = logging.Formatter(
        "%(asctime)s [%(levelname)s] %(name)s %(funcName)s:%(lineno)d %(message)s"
    )

    handler = logging.handlers.RotatingFileHandler(
        filename=path,
        maxBytes=10485760,  # 10 Megabytes
        backupCount=10,
    )
    handler.setLevel(level)
    handler.setFormatter(formatter)

    return handler


def print_changelog(version: str, cached_version_path: Optional[str]) -> None:
    """Prints the latest section of the CHANGELOG to stdout.

    This function prints the contents of the CHANGELOG (extracted from the PKG-INFO metadata)
    between the current version (`version`) and the latest cached version (extracted from the
    provided file path).

    Args:
        version: the currently running version of coBib.
        cached_version_path: the path to the file which caches the previously executed version of
            coBib. If `None`, this method exits early (thereby silencing this feature).
    """
    if cached_version_path is None:
        return

    try:
        current_version = version[: version.index("+")]
    except ValueError:
        current_version = version

    cached_version = None
    try:
        with open(RelPath(cached_version_path).path, "r", encoding="utf-8") as version_file:
            cached_version = version_file.read().strip()
    except FileNotFoundError:
        pass

    if current_version == cached_version:
        return

    with open(RelPath(cached_version_path).path, "w", encoding="utf-8") as version_file:
        version_file.write(current_version)

    lines = ["\x1b[1mHi there! It looks like you have updated coBib; here is what's new:\x1b[22m\n"]

    metadata = ""
    try:
        metadata = get_distribution("cobib").get_metadata("METADATA")
    except FileNotFoundError:
        try:
            metadata = get_distribution("cobib").get_metadata("PKG-INFO")
        except FileNotFoundError:
            lines += [
                "I wanted to show you the new changes here but was unable to query ",
                "them from your installation. You can look them up yourself, here: ",
                "https://gitlab.com/mrossinek/cobib/-/blob/master/CHANGELOG.md",
            ]

    num_printed_lines = -1
    for line in metadata.splitlines():
        line = line.strip()
        if line.startswith(f"## [{current_version}]"):
            num_printed_lines = 0
        elif line.startswith(f"## [{cached_version}]"):
            num_printed_lines = -1
        elif num_printed_lines >= 20:
            num_printed_lines = -1
            lines.extend(
                [
                    "\n...\n\x1b[31m",
                    "This output is shortened for the sake of brevity! For more information visit:",
                    "https://gitlab.com/mrossinek/cobib/-/blob/master/CHANGELOG.md",
                ]
            )

        if num_printed_lines >= 0:
            lines.append(line)
            num_printed_lines += 1

    print("\x1b[33m", end="")
    print("\n".join(lines).strip())
    print("\x1b[0m", end="", flush=True)

    input("\nPress Enter to continue...")



"""Mocking utilities for the `cobib.tui.tui.TUI`."""
# pylint: disable=missing-function-docstring,unused-argument

import logging
from typing import Dict, Set

from .mock_curses import MockCursesPad


class MockTUI:
    """This class mocks the `cobib.tui.tui.TUI`.

    When testing the `cobib.tui` modules, we need full control. Thus, this class removes all of the
    non-essential aspects and logs all method calls.
    """

    ANSI_MAP: Dict[str, int] = {}
    """The `ANSI_MAP` exists as part of `cobib.tui.tui.TUI` and will be populated at runtime."""

    def __init__(self) -> None:
        # noqa: D107
        self.height = 20
        self.width = 40
        self.prompt = MockCursesPad()
        self.selection: Set[str] = set()
        self.topbar = None
        self.logger = logging.getLogger("MockTUI")

    def prompt_handler(self, *args, **kwargs):  # type: ignore
        # noqa: D102
        self.logger.debug("prompt_handler")

    def resize_handler(self, *args):  # type: ignore
        # noqa: D102
        self.logger.debug("resize_handler")

    def statusbar(self, *args):  # type: ignore
        # noqa: D102
        self.logger.debug("statusbar")



"""coBib's YAML parser.

This parser leverages the [`ruamel.yaml`](https://pypi.org/project/ruamel.yaml/) library to convert
between `cobib.database.Entry` instances and YAML representations of their `dict`-like data
structure.

The parser is registered under the `-y` and `--yaml` command-line arguments of the
`cobib.commands.add.AddCommand`.

The following documentation is mostly inherited from the abstract interface
`cobib.parsers.base_parser`.
"""

import io
import logging
from collections import OrderedDict
from pathlib import Path
from typing import IO, Dict, Optional, Union, cast

from ruamel import yaml

from cobib.config import Event
from cobib.database import Entry

from .base_parser import Parser

LOGGER = logging.getLogger(__name__)


class YAMLParser(Parser):
    """The YAML Parser."""

    name = "yaml"

    class YamlDumper(yaml.YAML):  # type: ignore
        """Wrapper class for YAML dumping."""

        # pylint: disable=arguments-differ,inconsistent-return-statements
        def dump(self, data, stream=None, **kw) -> Optional[str]:  # type: ignore
            """A wrapper to dump as a string.

            Adapted from
            [here](https://yaml.readthedocs.io/en/latest/example.html#output-of-dump-as-a-string).
            """
            inefficient = False
            if stream is None:  # pragma: no branch
                inefficient = True
                stream = yaml.compat.StringIO()
            yaml.YAML.dump(self, data, stream, **kw)  # type: ignore
            if inefficient:
                return cast(str, stream.getvalue())
            return None  # pragma: no cover

    def parse(self, string: Union[str, Path]) -> Dict[str, Entry]:
        # pdoc will inherit the docstring from the base class
        # noqa: D102

        string = Event.PreYAMLParse.fire(string) or string

        bib = OrderedDict()
        LOGGER.debug("Loading YAML data from file: %s.", string)
        try:
            stream: IO = io.StringIO(Path(string))  # type: ignore
        except TypeError:
            try:
                stream = open(string, "r", encoding="utf-8")  # pylint: disable=consider-using-with
            except FileNotFoundError as exc:
                raise exc
        yml = yaml.YAML(typ="safe", pure=True)  # type: ignore
        for entry in yml.load_all(stream):
            for label, data in entry.items():
                bib[label] = Entry(label, data)
        stream.close()

        Event.PostYAMLParse.fire(bib)

        return bib

    def dump(self, entry: Entry) -> Optional[str]:
        # pdoc will inherit the docstring from the base class
        # noqa: D102

        Event.PreYAMLDump.fire(entry)

        yml = self.YamlDumper()
        yml.explicit_start = True
        yml.explicit_end = True
        LOGGER.debug("Converting entry %s to YAML format.", entry.label)
        string = yml.dump({entry.label: dict(sorted(entry.data.items()))})

        string = Event.PostYAMLDump.fire(string) or string

        return string



"""coBib's Open command.

This command can be used to open associated files of an entry.
```
cobib open <label 1> [<label 2> ...]
```

The `file` and `url` fields of `cobib.database.Entry.data` are queried for Path or URL strings.
If one such string is found, it is automatically opened with the program configured by
`config.commands.open.command`.
If multiple matches are found, the user will be presented with a menu to choose one or multiple
matches.

This menu will look similar to the following after querying for `help`:
```
You can specify one of the following options:
  1. a url number
  2. a field name provided in '[...]'
  3. or simply 'all'
  4. ENTER will abort the command

  1: [file] /path/to/a/file.pdf
  2: [file] /path/to/another/file.pdf
  3: [url] https://example.org/
Entry to open [Type 'help' for more info]:
```

With the above options, here is what will happen depending on the users choice:
* `1`, `2`, or `3`: will open the respective file or URL.
* `file` or `url`: will open the respective group.
* `all`: will open all matches.
* `help`: will print the detailed help-menu again.
* `ENTER`: will abort the command.

You can also trigger this command from the `cobib.tui.tui.TUI`.
By default, it is bound to the `o` key.
"""

from __future__ import annotations

import argparse
import logging
import os
import subprocess
import sys
from collections import defaultdict
from typing import IO, TYPE_CHECKING, Any, Dict, List
from urllib.parse import ParseResult, urlparse

from cobib.config import Event, config
from cobib.database import Database
from cobib.utils.rel_path import RelPath

from .base_command import ArgumentParser, Command

LOGGER = logging.getLogger(__name__)

if TYPE_CHECKING:
    import cobib.tui


class OpenCommand(Command):
    """The Open Command."""

    name = "open"

    # pylint: disable=too-many-branches
    def execute(self, args: List[str], out: IO[Any] = sys.stderr) -> None:
        """Opens associated files of an entry.

        This command opens the associated file(s) of one (or multiple) entries.
        It does so by querying the `file` and `url` fields of `cobib.database.Entry.data`.
        If multiple such files are found, the user is presented with a menu allowing him to choose
        one or multiple files to be opened.

        The command for opening can be configured via `config.commands.open.command`.

        Args:
            args: a sequence of additional arguments used for the execution. The following values
                are allowed for this command:
                    * `labels`: one (or multiple) labels of the entries to be opened.
            out: the output IO stream. This defaults to `sys.stdout`.
        """
        LOGGER.debug("Starting Open command.")
        parser = ArgumentParser(prog="open", description="Open subcommand parser.")
        parser.add_argument("labels", type=str, nargs="+", help="labels of the entries")

        if not args:
            parser.print_usage(sys.stderr)
            sys.exit(1)

        try:
            largs = parser.parse_args(args)
        except argparse.ArgumentError as exc:
            LOGGER.error(exc.message)
            return

        Event.PreOpenCommand.fire(largs)

        bib = Database()

        # pylint: disable=too-many-nested-blocks
        for label in largs.labels:
            things_to_open: Dict[str, List[ParseResult]] = defaultdict(list)
            count = 0
            # first: find all possible things to open
            try:
                entry = bib[label]
                for field in ("file", "url"):
                    if field in entry.data.keys() and entry.data[field]:
                        value = entry.data[field]
                        for val in value:
                            val = val.strip()
                            LOGGER.debug('Parsing "%s" for URLs.', val)
                            things_to_open[field] += [urlparse(val)]
                            count += 1
            except KeyError:
                msg = f"No entry with the label '{label}' could be found."
                LOGGER.warning(msg)
                continue

            # if there are none, skip current label
            if not things_to_open:
                msg = f"The entry '{label}' has no actionable field associated with it."
                LOGGER.warning(msg)
                continue

            if count == 1:
                # we found a single URL to open
                self._open_url(list(things_to_open.values())[0][0])
            else:
                # we query the user what to do
                idx = 1
                url_list = []
                prompt: List[str] = []
                # print formatted list of available URLs
                for field, urls in things_to_open.items():
                    for url in urls:
                        prompt.append(f"{idx:3}: [{field}] {url.geturl()}")
                        url_list.append(url)
                        idx += 1
                # loop until the user picks a valid choice
                help_requested = False
                while True:
                    prompt_copy = prompt.copy()
                    prompt_copy.append("Entry to open [Type 'help' for more info]: ")
                    try:
                        choice = input("\n".join(prompt_copy)).strip()
                    except EOFError:
                        choice = ""
                    if not choice:
                        # empty input
                        msg = "User aborted open command."
                        LOGGER.warning(msg)
                        break
                    if choice == "help":
                        LOGGER.debug("User requested help.")
                        if not help_requested:
                            prompt = [
                                "You can specify one of the following options:",
                                "  1. a url number",
                                "  2. a field name provided in '[...]'",
                                "  3. or simply 'all'",
                                "  4. ENTER will abort the command",
                                "",
                            ] + prompt
                        help_requested = True
                    elif choice == "all":
                        LOGGER.debug("User selected all urls.")
                        for url in url_list:
                            self._open_url(url)
                        break
                    elif choice in things_to_open.keys():
                        LOGGER.debug("User selected the %s set of urls.", choice)
                        for url in things_to_open[choice]:
                            self._open_url(url)
                        break
                    elif choice.isdigit() and int(choice) > 0 and int(choice) <= count:
                        LOGGER.debug("User selected url %s", choice)
                        self._open_url(url_list[int(choice) - 1])
                        break

        Event.PostOpenCommand.fire(largs.labels)

    @staticmethod
    def _open_url(url: ParseResult) -> None:
        """Opens a URL."""
        opener = config.commands.open.command
        try:
            url_str: str = url.geturl() if url.scheme else str(RelPath(url.geturl()).path)
            LOGGER.debug('Opening "%s" with %s.', url_str, opener)
            with open(os.devnull, "w", encoding="utf-8") as devnull:
                subprocess.Popen(  # pylint: disable=consider-using-with
                    [opener, url_str], stdout=devnull, stderr=devnull, stdin=devnull, close_fds=True
                )
        except FileNotFoundError as err:
            LOGGER.error(err)

    @staticmethod
    def tui(tui: cobib.tui.TUI) -> None:
        # pdoc will inherit the docstring from the base class
        # noqa: D102
        LOGGER.debug("Open command triggered from TUI.")
        if tui.selection:
            # use selection for command
            labels = list(tui.selection)
        else:
            # get current label
            label, _ = tui.viewport.get_current_label()
            labels = [label]
        tui.execute_command(["open"] + labels, skip_prompt=True)



"""coBib's Command interface."""

from __future__ import annotations

import argparse
import json
import logging
import os
import shlex
import sys
from abc import ABC, abstractmethod
from typing import IO, TYPE_CHECKING, Any, Dict, Generator, List, NoReturn, Optional

from cobib.config import Event, config
from cobib.utils.rel_path import RelPath

LOGGER = logging.getLogger(__name__)

if TYPE_CHECKING:
    import cobib.tui


class Command(ABC):
    """The Command interface.

    This interface should be implemented by all concrete command implementations.
    In order to provide both, command-line and TUI usability, the `execute` and `tui` functions must
    be implemented, respectively.
    """

    name = "base"
    """The commands `name` is used to extract the available commands for the command-line interface.
    """

    def __init__(self) -> None:
        """The initializer of any concrete implementation should *not* take any arguments!"""

    @abstractmethod
    def execute(self, args: List[str], out: IO[Any] = sys.stdout) -> Any:
        """Actually executes the command.

        This means, all of the command-specific logic and action needs to be implemented by this
        method.
        It also poses as the pure interface triggered from the command-line.
        The arguments given to this function are parsed by `argparse` which also means that each
        subcommand should provide an additional `--help` menu for the command-line interface.
        ```
        cobib <subcommand> --help
        ```

        This function may *not* raise any errors!
        This also means it should take care of catching all potential errors triggered by internally
        used methods.
        Such encountered errors should be logged and the method should return gracefully.

        Args:
            args: a sequence of additional arguments used for the execution.
            out: the output IO stream. This defaults to `sys.stdout`.

        Returns:
            Usually `None` but some complex commands may choose to return some of their runtime data
            for ease of additional post-processing in (e.g.) the `tui` wrapper.
        """

    @staticmethod
    @abstractmethod
    def tui(tui: cobib.tui.TUI) -> Optional[Generator[List[str], None, None]]:
        """TUI command interface.

        This function serves as the commands entry-point from the `cobib.tui.tui.TUI` instance.
        It should take care of any pre- and/or post-processing before calling `execute` internally
        to do the actual work.

        The processing may involve handling of highlighting as well as general screen buffer
        contents.

        Args:
            tui: the runtime-instance of coBib's TUI.

        Yields:
            Optionally, this method may `yield` the command arguments for further processing.
        """

    def git(self, args: Optional[Dict[str, Any]] = None, force: bool = False) -> None:
        """Generates a git commit to track the commands changes.

        This function only has an effect when `config.database.git` is enabled *and* the database
        has been initialized correctly with `cobib.commands.init.InitCommand`.
        Otherwise, a warning will be printed and no commit will be generated.
        Nonetheless, the changes applied by the commit will have taken effect in the database.

        Args:
            args: a dictionary containing the *parsed* command arguments.
            force: whether to ignore the configuration setting. This option is mainly used by the
                `cobib.commands.init.InitCommand`.
        """
        git_tracked = config.database.git
        if not git_tracked and not force:
            return

        file = RelPath(config.database.file).path
        root = file.parent

        if not (root / ".git").exists():
            if git_tracked:
                msg = (
                    "You have configured coBib to track your database with git."
                    "\nPlease run `cobib init --git`, to initialize this tracking."
                )
                LOGGER.warning(msg)
                return

        msg = f"Auto-commit: {self.name.title()}Command"
        if args:
            msg += "\n\n"
            msg += json.dumps(args, indent=2, default=str)

        msg = Event.PreGitCommit.fire(msg, args) or msg

        commands = [
            f"cd {root}",
            f"git add -- {file}",
            f"git commit --no-gpg-sign --quiet --message {shlex.quote(msg)}",
        ]
        LOGGER.debug("Auto-commit to git from %s command.", self.name)
        os.system("; ".join(commands))

        Event.PostGitCommit.fire(root, file)


class ArgumentParser(argparse.ArgumentParser):
    """Wrapper of the `argparse.ArgumentParser` to allow catching of error messages.

    Note, this class will be removed once Python 3.9 becomes the minimal supported version as it
    added the [`exit_on_error`](https://docs.python.org/3/library/argparse.html#exit-on-error)
    keyword argument.
    """

    # TODO: once Python 3.9 becomes the default, make use of the exit_on_error argument.

    def exit(self, status: int = 0, message: Optional[str] = None) -> NoReturn:
        """Overwrite the exit method to raise an error rather than exit.

        Args:
            status: the status code. If non-zero, an `argparse.ArgumentError` will be raised.
            message: the message of the error.

        Raises:
            An `argparse.ArgumentError`.
        """
        if status:
            raise argparse.ArgumentError(None, f"Error: {message}")
        super().exit(status, message)  # pragma: no cover



"""Tests for coBib's UndoCommand."""
# pylint: disable=no-self-use,unused-argument

from __future__ import annotations

import contextlib
import logging
import os
import subprocess
from argparse import Namespace
from io import StringIO
from pathlib import Path
from shutil import rmtree
from typing import TYPE_CHECKING, Any, Type

import pytest

from cobib.commands import AddCommand, UndoCommand
from cobib.config import Event, config
from cobib.database import Database

from .. import get_resource
from ..tui.tui_test import TUITest
from .command_test import CommandTest

EXAMPLE_MULTI_FILE_ENTRY_BIB = get_resource("example_multi_file_entry.bib", "commands")

if TYPE_CHECKING:
    import cobib.commands


class TestUndoCommand(CommandTest, TUITest):
    """Tests for coBib's UndoCommand."""

    def get_command(self) -> Type[cobib.commands.base_command.Command]:
        # noqa: D102
        return UndoCommand

    def _assert(self) -> None:
        """Common assertion utility method."""
        assert Database().get("example_multi_file_entry", None) is None

        # get last commit message
        with subprocess.Popen(
            ["git", "-C", self.COBIB_TEST_DIR, "show", "--format=format:%B", "--no-patch", "HEAD"],
            stdout=subprocess.PIPE,
        ) as proc:
            message, _ = proc.communicate()
            # decode it
            split_message = message.decode("utf-8").split("\n")
            # assert subject line
            assert "Undo" in split_message[0]

    @pytest.mark.parametrize(
        ["setup", "expected_exit"],
        [
            [{"git": False}, False],
            [{"git": True}, False],
            [{"git": True}, True],
        ],
        indirect=["setup"],
    )
    def test_command(
        self, setup: Any, expected_exit: bool, caplog: pytest.LogCaptureFixture
    ) -> None:
        """Test the command itself.

        Args:
            setup: the `tests.commands.command_test.CommandTest.setup` fixture.
            expected_exit: whether to expect an early exit.
            caplog: the built-in pytest fixture.
        """
        git = setup.get("git", False)

        if not git:
            UndoCommand().execute([])
            for (source, level, message) in caplog.record_tuples:
                if ("cobib.commands.undo", logging.ERROR) == (
                    source,
                    level,
                ) and "git-tracking" in message:
                    break
            else:
                pytest.fail("No Error logged from UndoCommand.")
        elif expected_exit:
            # Regression test related to #65
            with pytest.raises(SystemExit):
                UndoCommand().execute([])
            for (source, level, message) in caplog.record_tuples:
                if ("cobib.commands.undo", logging.WARNING) == (
                    source,
                    level,
                ) and "Could not find a commit to undo." in message:
                    break
            else:
                pytest.fail("No Error logged from UndoCommand.")
        else:
            AddCommand().execute(["-b", EXAMPLE_MULTI_FILE_ENTRY_BIB])
            UndoCommand().execute([])

            self._assert()

    @pytest.mark.parametrize(
        ["setup"],
        [
            [{"git": True}],
        ],
        indirect=["setup"],
    )
    def test_skipping_undone_commits(self, setup: Any, caplog: pytest.LogCaptureFixture) -> None:
        """Test skipping already undone commits.

        Args:
            setup: the `tests.commands.command_test.CommandTest.setup` fixture.
            caplog: the built-in pytest fixture.
        """
        AddCommand().execute(["-b", EXAMPLE_MULTI_FILE_ENTRY_BIB])
        AddCommand().execute(["-b", get_resource("example_entry.bib")])
        UndoCommand().execute([])
        caplog.clear()

        UndoCommand().execute([])
        self._assert()
        assert "Storing undone commit" in caplog.record_tuples[4][2]
        assert "Skipping" in caplog.record_tuples[6][2]

    @pytest.mark.parametrize(
        ["setup"],
        [
            [{"git": True}],
        ],
        indirect=["setup"],
    )
    def test_warn_insufficient_setup(self, setup: Any, caplog: pytest.LogCaptureFixture) -> None:
        """Test warning in case of insufficient setup.

        Args:
            setup: the `tests.commands.command_test.CommandTest.setup` fixture.
            caplog: the built-in pytest fixture.
        """
        rmtree(self.COBIB_TEST_DIR_GIT)
        UndoCommand().execute([])
        for (source, level, message) in caplog.record_tuples:
            if ("cobib.commands.undo", logging.ERROR) == (
                source,
                level,
            ) and "configured, but not initialized" in message:
                break
        else:
            pytest.fail("No Error logged from UndoCommand.")

    @pytest.mark.parametrize(
        ["setup"],
        [
            [{"git": True}],
        ],
        indirect=["setup"],
    )
    # other variants are already covered by test_command
    def test_cmdline(
        self, setup: Any, monkeypatch: pytest.MonkeyPatch, caplog: pytest.LogCaptureFixture
    ) -> None:
        """Test the command-line access of the command.

        Args:
            setup: the `tests.commands.command_test.CommandTest.setup` fixture.
            monkeypatch: the built-in pytest fixture.
            caplog: the built-in pytest fixture.
        """
        AddCommand().execute(["-b", EXAMPLE_MULTI_FILE_ENTRY_BIB])
        self.run_module(monkeypatch, "main", ["cobib", "undo"])

        self._assert()

    # manually overwrite this test because we must enable git integration
    def test_handle_argument_error(self, caplog: pytest.LogCaptureFixture) -> None:
        """Test handling of ArgumentError.

        Args:
            caplog: the built-in pytest fixture.
        """
        # use temporary config
        config.database.file = self.COBIB_TEST_DIR / "database.yaml"
        config.database.git = True

        # initialize git-tracking
        self.COBIB_TEST_DIR.mkdir(parents=True, exist_ok=True)
        open(  # pylint: disable=consider-using-with
            config.database.file, "w", encoding="utf-8"
        ).close()
        os.system("git init " + str(self.COBIB_TEST_DIR))

        try:
            super().test_handle_argument_error(caplog)
        finally:
            # clean up file system
            rmtree(self.COBIB_TEST_DIR_GIT)
            # clean up config
            config.defaults()

    @pytest.mark.parametrize(
        ["setup"],
        [
            [{"git": True}],
        ],
        indirect=["setup"],
    )
    def test_tui(self, setup: Any) -> None:
        """Test the TUI access of the command.

        Args:
            setup: the `tests.commands.command_test.CommandTest.setup` fixture.
        """

        def assertion(screen, logs, **kwargs):  # type: ignore
            # check that the undone entry has actually been present in the buffer before
            assert (
                "cobib.tui.buffer",
                10,
                "Appending string to text buffer: example_multi_file_entry",
            ) in logs
            # but is no longer there, now
            assert "example_multi_file_entry" not in screen.display[1]

            expected_log = [
                ("cobib.commands.undo", 10, "Undo command triggered from TUI."),
                ("cobib.commands.undo", 10, "Starting Undo command."),
                ("cobib.commands.undo", 10, "Obtaining git log."),
            ]
            # we only assert the first three messages because the following ones will contain always
            # changing commit SHAs
            assert [log for log in logs if log[0] == "cobib.commands.undo"][0:3] == expected_log

        AddCommand().execute(["-b", EXAMPLE_MULTI_FILE_ENTRY_BIB])
        self.run_tui("u", assertion, {})

    @pytest.mark.parametrize("setup", [{"git": True}], indirect=["setup"])
    def test_event_pre_undo_command(self, setup: Any) -> None:
        """Tests the PreUndoCommand event."""

        @Event.PreUndoCommand.subscribe
        def hook(largs: Namespace) -> None:
            print("Hello world!")

        assert Event.PreUndoCommand.validate()

        with contextlib.redirect_stdout(StringIO()) as out:
            with pytest.raises(SystemExit):
                UndoCommand().execute([])

            assert out.getvalue() == "Hello world!\n"

    @pytest.mark.parametrize("setup", [{"git": True}], indirect=["setup"])
    def test_event_post_undo_command(self, setup: Any) -> None:
        """Tests the PostUndoCommand event."""

        @Event.PostUndoCommand.subscribe
        def hook(root: Path, sha: str) -> None:
            print(root)

        assert Event.PostUndoCommand.validate()

        with contextlib.redirect_stdout(StringIO()) as out:
            AddCommand().execute(["-b", EXAMPLE_MULTI_FILE_ENTRY_BIB])
            UndoCommand().execute([])

            self._assert()

            assert out.getvalue() == f"{self.COBIB_TEST_DIR}\n"



"""coBib's file downloader utility."""

from __future__ import annotations

import logging
import re
import sys
import tempfile
from typing import Callable, Optional

import requests

from cobib.config import Event, config

from .rel_path import RelPath

LOGGER = logging.getLogger(__name__)


class FileDownloader:
    """The file downloader singleton.

    This utility centralizes the downloading of associated files. It implements the singleton
    pattern to allow simple log method replacement (via `set_logger`).
    """

    _instance: Optional[FileDownloader] = None
    """The singleton instance of this class."""

    _logger: Callable[[str], None] = lambda text: print(text, end="", flush=True, file=sys.stdout)
    """The logging method used to display the downloading progress bar."""

    def __new__(cls) -> FileDownloader:
        """Singleton constructor.

        This method gets called when accessing `FileDownloader` and enforces the singleton pattern
        implemented by this class.
        """
        if cls._instance is None:
            cls._instance = super().__new__(cls)
        return cls._instance

    @staticmethod
    def set_logger(log_method: Callable[[str], None]) -> None:
        """Sets the class-wide logging method (see also `FileDownloader._logger`).

        This method is used to display the progress bar of the file downloading.

        Args:
            log_method: the logging method.
        """
        FileDownloader._logger = log_method

    # bytes pretty-printing
    _UNITS_MAPPING = [
        (1 << 50, " PB"),
        (1 << 40, " TB"),
        (1 << 30, " GB"),
        (1 << 20, " MB"),
        (1 << 10, " KB"),
        (1, " B"),
    ]
    """Maps byte sizes to units."""

    @staticmethod
    def _size(bytes_: int) -> str:
        """Human-readable file size.

        Args:
            bytes_: the size in bytes.

        Returns:
            The size formatted for easy human readability.

        Reference:
            https://stackoverflow.com/a/12912296
        """
        for factor, suffix in FileDownloader._UNITS_MAPPING:
            if bytes_ >= factor:
                break
        amount = int(bytes_ / factor)
        return str(amount) + suffix

    _PDF_MARKER = bytes("%PDF", "utf-8")
    """A marker which the downloaded file's beginning is checked against, to determine that it is
    indeed a PDF file."""

    @staticmethod
    def _assert_pdf(content: bytes) -> bool:
        """Asserts that the `content` starts with the `_PDF_MARKER`.

        Args:
            content: the string of bytes to check.

        Returns:
            Whether the`content` matches.
        """
        if not content.startswith(FileDownloader._PDF_MARKER):
            LOGGER.warning("The URL did not provide a PDF file. Aborting download!")
            return False
        return True

    @staticmethod
    def _unlink(path: RelPath) -> None:
        """Remove a file and ignore any error.

        Args:
            path: the file to remove.
        """
        try:
            # TODO: once Python 3.7 is dropped, leverage `missing_ok` argument
            path.path.unlink()
        except FileNotFoundError:
            pass

    def download(
        self, url: str, label: str, folder: Optional[str] = None, overwrite: bool = False
    ) -> Optional[RelPath]:
        """Downloads a file.

        The path of the downloaded file is `folder/label.pdf`. The path can be configured via
        `cobib.config.commands.add.download_location`.

        Args:
            url: the link to the file to be downloaded.
            label: the name of the entry.
            folder: an optional folder where the downloaded file will be stored.
            overwrite: whether or not to overwrite an existing file.

        Returns:
            The `RelPath` to the downloaded file. If downloading was not successful, `None` is
            returned.
        """
        if folder is None:
            folder = config.utils.file_downloader.default_location

        hook_result = Event.PreFileDownload.fire(url, label, folder)
        if hook_result is not None:
            url, label, folder = hook_result

        path = RelPath(f"{folder}/{label}.pdf")

        backup = None
        if path.path.exists():
            if not overwrite:
                LOGGER.warning(
                    "A file at '%s' already exists! Using that rather than downloading.", path
                )
                return path
            # we make a copy of the existing file in case downloading a new one fails
            backup = tempfile.NamedTemporaryFile()  # pylint: disable=consider-using-with
            backup.write(path.path.read_bytes())
            backup.seek(0)

        for pattern_url, repl_url in config.utils.file_downloader.url_map.items():
            if re.match(pattern_url, url):
                new_url = re.sub(pattern_url, repl_url, url)
                LOGGER.info(
                    "Matched the file's URL to your pattern URL %s and replaced it to become %s",
                    pattern_url,
                    new_url,
                )
                url = new_url
                break

        with open(path.path, "wb") as file:
            LOGGER.info("Downloading %s to %s", url, path)
            try:
                response = requests.get(url, timeout=10, stream=True)
                total_length = int(response.headers.get("content-length", -1))
            except requests.exceptions.RequestException as err:
                msg = f"An Exception occurred while downloading the file located at {url}"
                LOGGER.warning(msg)
                LOGGER.error(err)
                FileDownloader._unlink(path)
                if backup is not None:
                    path.path.write_bytes(backup.read())
                    backup.close()
                return None
            if total_length < 0:
                if not FileDownloader._assert_pdf(response.content):
                    FileDownloader._unlink(path)
                    if backup is not None:
                        path.path.write_bytes(backup.read())
                        backup.close()
                    return None
                file.write(response.content)
            else:
                accumulated_length = 0
                total_size = self._size(total_length)
                for data in response.iter_content(chunk_size=4096):
                    if accumulated_length == 0 and not FileDownloader._assert_pdf(data):
                        FileDownloader._unlink(path)
                        if backup is not None:
                            path.path.write_bytes(backup.read())
                            backup.close()
                        return None
                    accumulated_length += len(data)
                    file.write(data)
                    percentage = accumulated_length / total_length
                    progress = int(40 * percentage)
                    FileDownloader._logger(
                        "\rDownloading:"
                        f" [{'=' * progress}{' ' * (40 - progress)}] "
                        f"{100*percentage:6.1f}%"
                        f"{self._size(accumulated_length): >7} / {total_size: <7}",
                    )
                FileDownloader._logger("\n")
            msg = f"Successfully downloaded {path}"
            print(msg)
            LOGGER.info(msg)

            path = Event.PostFileDownload.fire(path) or path

            return path
